from fileformats.core import Classifier, DataType, __version__
from fileformats.core.mixin import WithClassifiers


class SubpackageClassified(WithClassifiers, DataType):
    classifiers_attr_name = "classifiers"
    classifiers = ()
    allowed_classifiers = (Classifier,)
    generically_classifiable = False


class Psi(DataType):
    pass


class Zeta(Classifier):
    pass


class Theta(Classifier):
    pass


__all__ = ["SubpackageClassified", "Psi", "Zeta", "Theta", "__version__"]



"""Adds support for "field" data types

While not "file-formats", in some workflows, particularly when interacting with a data
store such as in Arcana, it is useful to interchangeably interact with file and field
data, so the classes in this module are provided to support these use cases.
"""

import decimal
import typing as ty
from fileformats.core import Field, __version__  # noqa: F401
from fileformats.core.mixin import WithClassifier
from fileformats.core.exceptions import FormatMismatchError

ValueType = ty.TypeVar("ValueType")
PrimitiveType = ty.TypeVar("PrimitiveType")


class Singular(Field[ValueType, PrimitiveType]):
    pass


class LogicalMixin:
    def __bool__(self) -> bool:
        return bool(self.value)  # type: ignore

    def __and__(self, other: "LogicalMixin") -> bool:
        return self.value and other.value  # type: ignore

    def __or__(self, other: "LogicalMixin") -> bool:
        return self.value or other.value  # type: ignore

    def __not__(self) -> bool:
        return not self.value  # type: ignore

    def __invert__(self) -> bool:
        return not self.value  # type: ignore


class ScalarMixin(LogicalMixin, Field[ValueType, PrimitiveType]):
    value: ValueType

    def __add__(self, other: PrimitiveType) -> PrimitiveType:
        return self.value + other  # type: ignore

    def __sub__(self, other: PrimitiveType) -> PrimitiveType:
        return self.value - other  # type: ignore

    def __mul__(self, other: PrimitiveType) -> PrimitiveType:
        return self.value * other  # type: ignore

    def __truediv__(self, other: PrimitiveType) -> PrimitiveType:
        return self.value / other  # type: ignore

    def __floordiv__(self, other: PrimitiveType) -> PrimitiveType:
        return self.value // other  # type: ignore

    def __mod__(self, other: PrimitiveType) -> PrimitiveType:
        return self.value % other  # type: ignore

    def __pow__(self, other: PrimitiveType) -> PrimitiveType:
        return self.value**other  # type: ignore

    def __neg__(self) -> PrimitiveType:
        return -self.value  # type: ignore

    def __pos__(self) -> PrimitiveType:
        return +self.value  # type: ignore

    def __abs__(self) -> PrimitiveType:
        return abs(self.value)  # type: ignore


class Text(Singular[str, str]):
    value: str

    primitive = str

    def __init__(self, value: ty.Any):
        try:
            self.value = str(value)
        except ValueError as e:
            raise FormatMismatchError(str(e)) from None

    def __hash__(self) -> int:
        return hash(self.value)

    def __repr__(self) -> str:
        return f'{self.type_name}("{self.value}")'


class Integer(Singular[int, int], ScalarMixin[int, int]):
    value: int

    primitive = int

    def __init__(self, value: ty.Any):
        if isinstance(value, float):
            raise FormatMismatchError(
                f"Cannot convert float ({value}) to integer field without potential loss "
                "of information"
            )
        try:
            self.value = int(value)
        except ValueError as e:
            raise FormatMismatchError(str(e)) from None

    def __int__(self) -> int:
        return self.value

    def __float__(self) -> float:
        return float(self.value)

    def __bool__(self) -> bool:
        return bool(self.value)

    def __hash__(self) -> int:
        return hash(self.value)


class Decimal(Singular[decimal.Decimal, float], ScalarMixin[decimal.Decimal, float]):
    value: decimal.Decimal

    primitive = float

    def __init__(self, value: ty.Any):
        if isinstance(value, Decimal):
            self.value = value.value
        try:
            self.value = (
                value.value if isinstance(value, Decimal) else decimal.Decimal(value)
            )
        except decimal.InvalidOperation as e:
            raise FormatMismatchError(str(e)) from None

    def __float__(self) -> float:
        return float(self.value)

    def __bool__(self) -> bool:
        return bool(self.value)

    def __hash__(self) -> int:
        return hash(self.value)


class Boolean(Singular[bool, bool], LogicalMixin):
    primitive = bool

    value: bool

    def __init__(self, value: ty.Any):
        if isinstance(value, str):
            if value.lower() in ("true", "1", "yes"):
                value = True
            elif value.lower() in ("false", "0", "no"):
                value = False
            else:
                raise FormatMismatchError(
                    f"Cannot convert string '{value}' to boolean value"
                )
        else:
            try:
                value = bool(value)
            except ValueError as e:
                raise FormatMismatchError(str(e)) from None
        assert isinstance(value, bool)
        self.value = value

    def __str__(self) -> str:
        return str(self.value).lower()

    def __bool__(self) -> bool:
        return self.value

    def __hash__(self) -> int:
        return hash(self.value)


ItemType = ty.TypeVar("ItemType", decimal.Decimal, int, float, bool)


class Array(
    WithClassifier,
    Field[ty.Tuple[ItemType, ...], ty.Tuple[ItemType, ...]],
    ty.Sequence[ItemType],
):
    # WithClassifiers class attrs
    classifiers_attr_name: str = "item_type"
    allowed_classifiers: ty.Tuple[
        ty.Type[Singular[ty.Tuple[ItemType, ...], ty.Tuple[ItemType, ...]]]
    ] = (Singular,)
    item_type: ty.Optional[ty.Type[Singular[ItemType, ty.Any]]] = None

    primitive = tuple

    value: ty.Tuple[ItemType, ...]

    def __init__(self, value: ty.Union[str, ty.Sequence[ty.Any]]):

        if isinstance(value, str):
            if (value.startswith("[") and value.endswith("]")) or (
                value.startswith("(") and value.endswith(")")
            ):
                value = value[1:-1]
            elif (
                value.startswith("[")
                or value.endswith("]")
                or value.startswith("(")
                or value.endswith(")")
            ):
                raise FormatMismatchError(f"Unmatched brackets in array field {value}")
            value = tuple(v.strip() for v in value.split(","))
        else:
            try:
                value = tuple(value)
            except ValueError as e:
                raise FormatMismatchError(str(e)) from None
        assert isinstance(value, tuple)
        # Ensure items are of the correct type
        # if self.item_type is not None:
        if self.item_type is not None:
            parsed_value: ty.Tuple[ItemType, ...] = tuple(
                self.item_type(i).value for i in value
            )
        else:
            parsed_value = value
        self.value = parsed_value

    def __str__(self) -> str:
        return (
            "["
            + ",".join(
                (
                    str(self.item_type(i)) if self.item_type is not None else i  # type: ignore
                )  # pylint: disable=not-callable
                for i in self.value
            )
            + "]"
        )

    def __iter__(self) -> ty.Iterator[ty.Any]:
        return iter(self.value)

    def __hash__(self) -> int:
        return hash(self.value)

    def __len__(self) -> int:
        return len(self.value)

    @ty.overload
    def __getitem__(self, index: int) -> ItemType:
        ...  # noqa: E704

    @ty.overload
    def __getitem__(self, slice: slice) -> ty.Sequence[ItemType]:
        ...  # noqa: E704

    def __getitem__(self, key: ty.Any) -> ty.Any:
        return self.value[key]



import json
import typing as ty
from fileformats.core.typing import TypeAlias
from pathlib import Path
from fileformats.core import DataType, FileSet, extra_implementation
from fileformats.core.mixin import WithClassifier
from fileformats.generic import UnicodeFile
from fileformats.core.exceptions import FormatMismatchError
from fileformats.core import SampleFileGenerator


SerializationType: TypeAlias = ty.Union[ty.Dict[str, ty.Any], ty.List[ty.Any]]


class Schema(DataType):
    """Base class for all serialization schemas (can be used for abstract schemas that
    don't actually have a formal definition)"""

    pass


class JsonSchema(Schema):
    pass


class XmlSchema(Schema):
    pass


class InformalSchema(Schema):
    "Not actually a strict schema, just a set of conventions on how to structure the serialization"


class TextSerialization(WithClassifier, UnicodeFile):
    "Base class for text-based hierarchical data-serialization formats, e.g. JSON, YAML"

    # Classifiers class attrs
    classifiers_attr_name: ty.Optional[str] = "schema"
    schema: ty.Optional[ty.Type[Schema]] = None
    allowed_classifiers: ty.Tuple[ty.Type[Schema], ...] = (Schema,)
    generically_classifiable: bool = True


class Xml(TextSerialization):
    ext: ty.Optional[str] = ".xml"
    allowed_classifiers = (XmlSchema, InformalSchema)


class Json(TextSerialization):
    ext: ty.Optional[str] = ".json"
    allowed_classifiers = (JsonSchema, InformalSchema)

    # TODO: add validation mechanisms to check class


class Yaml(TextSerialization):
    ext = ".yaml"
    alternate_exts = (".yml",)


class Toml(TextSerialization):
    ext = ".toml"


@extra_implementation(FileSet.generate_sample_data)
def generate_json_sample_data(
    js: Json,
    generator: SampleFileGenerator,
) -> ty.List[Path]:
    js_file = generator.generate_fspath(file_type=Json)
    with open(js_file, "w") as f:
        json.dump(
            {
                "a": True,
                "b": "two",
                "c": 3,
                "d": [generator.rng.randint(0, 10), generator.rng.random(), 6],
            },
            f,
        )
    return [js_file]


@extra_implementation(FileSet.generate_sample_data)
def generate_yaml_sample_data(
    yml: Yaml,
    generator: SampleFileGenerator,
) -> ty.List[Path]:
    yml_file = generator.generate_fspath(file_type=Yaml)
    with open(yml_file, "w") as f:
        f.write(
            f"""# Generated sample YAML file by FileFormats
a: True
b: two
c: 3
d:
- {generator.rng.randint(0, 10)}
- {generator.rng.random()}
- 6
"""
        )
    return [yml_file]


@extra_implementation(FileSet.load)
def load(jsn: Json, **kwargs: ty.Any) -> SerializationType:
    try:
        with jsn.open() as f:
            dct: ty.Dict[str, ty.Any] = json.load(f, **kwargs)
    except json.JSONDecodeError as e:
        raise FormatMismatchError(f"'{jsn.fspath}' is not a valid JSON file") from e
    return dct


@extra_implementation(FileSet.save)
def save(jsn: Json, data: SerializationType, **kwargs: ty.Any) -> None:
    with jsn.open("w") as f:
        json.dump(data, f, **kwargs)



import importlib
from pathlib import Path
import inspect
import typing as ty
from types import ModuleType
import urllib.request
import urllib.error
import os
import logging
import pkgutil
from contextlib import contextmanager
from .typing import FspathsInputType
import fileformats.core
from fileformats.core.exceptions import FormatDefinitionError


logger = logging.getLogger("fileformats")


_excluded_subpackages = set(
    ["core", "testing", "serialization", "archive", "document", "conftest"]
)

T = ty.TypeVar("T")


def include_testing_package(flag: bool = True) -> None:
    """Include testing package in list of sub-packages. Typically set in conftest.py
    or similar when setting up unittesting. Must be set globally before any methods are
    called within the package as member classes are cached.

    Parameters
    ----------
    flag : bool
        whether to include the testing package or not
    """
    global _excluded_subpackages
    if flag:
        _excluded_subpackages.remove("testing")
    else:
        _excluded_subpackages.add("testing")


def subpackages(
    exclude: ty.Iterable[str] = _excluded_subpackages,
) -> ty.Generator[ModuleType, None, None]:
    """Iterates over all subpackages within the fileformats namespace

    Parameters
    ----------
    exclude : ty.Sequence[str], optional
        whether to include the testing subpackage, by default ["core", "testing"]

    Yields
    ------
    module
        all modules within the package
    """
    for mod_info in pkgutil.iter_modules(
        fileformats.__path__, prefix=fileformats.__package__ + "."
    ):
        if mod_info.name.split(".")[-1] in exclude:
            continue
        yield importlib.import_module(mod_info.name)


@contextmanager
def set_cwd(path: Path) -> ty.Generator[Path, None, None]:
    """Sets the current working directory to `path` and back to original
    working directory on exit

    Parameters
    ----------
    path : str
        The file system path to set as the current working directory
    """
    pwd = os.getcwd()
    os.chdir(path)
    try:
        yield path
    finally:
        os.chdir(pwd)


def fspaths_converter(fspaths: FspathsInputType) -> ty.FrozenSet[Path]:
    """Ensures fs-paths are a set of pathlib.Path"""
    import fileformats.core

    if isinstance(fspaths, fileformats.core.FileSet):
        fspaths = fspaths.fspaths
    elif isinstance(fspaths, (str, os.PathLike)):
        fspaths = [Path(fspaths)]
    return frozenset(Path(p).absolute() for p in fspaths)


def add_exc_note(e: Exception, note: str) -> Exception:
    """Adds a note to an exception in a Python <3.11 compatible way

    Parameters
    ----------
    e : Exception
        the exception to add the note to
    note : str
        the note to add

    Returns
    -------
    Exception
        returns the exception again
    """
    if hasattr(e, "add_note"):
        e.add_note(note)
    else:
        e.args = (e.args[0] + "\n" + note,)
    return e


def matching_source(
    task1: ty.Callable[..., ty.Any], task2: ty.Callable[..., ty.Any]
) -> bool:
    """Checks to see if the tasks share the same source code but are just getting reimported
    for some unknown reason"""
    mod1 = inspect.getmodule(task1)
    mod2 = inspect.getmodule(task2)
    assert mod1 and mod2
    return inspect.getsource(mod1) == inspect.getsource(mod2)


def check_package_exists_on_pypi(package_name: str, timeout: int = 5) -> bool:
    """Check if a package exists on PyPI

    Parameters
    ----------
    package_name : str
        the name of the package to check for

    Returns
    -------
    bool
        whether the package exists on PyPI or not
    """
    url = f"https://pypi.org/pypi/{package_name}/json"
    try:
        urllib.request.urlopen(url, timeout=timeout)
    except urllib.error.HTTPError as e:
        if e.code == 404:
            return False
        else:
            raise
    return True


class ExtrasModule:
    def __init__(self, imported: bool, pkg: ty.Optional[str], pypi: ty.Optional[str]):
        self.imported = imported
        self.pkg = pkg
        self.pypi = pypi

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}(pkg={self.pkg}, pypi={self.pypi}, "
            f"imported={self.imported})"
        )


def import_extras_module(klass: ty.Type["fileformats.core.DataType"]) -> ExtrasModule:
    """Attempt to load extras module corresponding to the provided class's module

    Parameters
    ----------
    klass : ty.Type
        the class to load the extras module for

    Returns
    -------
    imported : bool
        whether the module was imported or not
    sub_pkg : str
        the name of the sub-package that was attempted to be loaded
    """
    from .identification import IANA_MIME_TYPE_REGISTRIES

    # Check for Mock class
    try:
        klass = klass.TRUE_CLASS  # type: ignore
    except AttributeError:
        pass
    pkg_parts = klass.__module__.split(".")
    if pkg_parts[0] != "fileformats":
        logger.debug(
            "There is no 'extras' module for classes not within the 'fileformats' package, "
            "not %s in %s",
            klass.__name__,
            klass.__module__,
        )
        return ExtrasModule(True, None, None)
    sub_pkg = pkg_parts[1]
    extras_pkg = "fileformats.extras." + sub_pkg
    if sub_pkg in IANA_MIME_TYPE_REGISTRIES + ["testing"]:
        extras_pypi = "fileformats-extras"
    else:
        extras_pypi = f"fileformats-{sub_pkg.replace('_', '-')}-extras"
    try:
        importlib.import_module(extras_pkg)
    except ModuleNotFoundError as e:
        if str(e) != f"No module named '{extras_pkg}'":
            raise
        extras_imported = False
    else:
        extras_imported = True
    return ExtrasModule(extras_imported, extras_pkg, extras_pypi)


TypeType = ty.TypeVar("TypeType", bound=ty.Type[ty.Any])


def get_optional_type(
    type_: ty.Union[TypeType, ty.Type[ty.Optional[TypeType]]], allowed: bool = True
) -> TypeType:
    """Checks if a type is an Optional type

    Parameters
    ----------
    type_ : ty.Type
        the type to check
    allowed : bool
        whether Optional types are allowed or not

    Returns
    -------
    bool
        whether the type is an Optional type or not
    """
    if ty.get_origin(type_) is None:
        return type_  # type: ignore[return-value]
    if not allowed:
        raise FormatDefinitionError(
            f"Optional types are not allowed in content_type definitions ({type_}) "
            "in this context"
        )
    args = ty.get_args(type_)
    if len(args) != 2 and None in ty.get_args(type_):
        raise FormatDefinitionError(
            "Only Optional types are allowed in content_type definitions, "
            f"not {type_}"
        )
    return args[0] if args[0] is not None else args[1]  # type: ignore[no-any-return]



import typing as ty
from fileformats.core import FileSet
from fileformats.core.mixin import WithClassifier
from fileformats.generic import BinaryFile
from fileformats.core.mixin import WithMagicNumber


class Archive(BinaryFile):
    "Base class for compressed archives"
    archived_type: ty.Optional[ty.Type[FileSet]] = None


class WithArchiveClassifiers(WithClassifier):
    "Base class for compressed archives"

    classifiers_attr_name = "archived_type"
    allowed_classifiers = (FileSet,)
    generically_classifiable = True


# Compressed formats
class BaseZip(WithMagicNumber, Archive):
    ext = ".zip"
    magic_number = "504B0304"
    iana_mime = "application/zip"


class BaseBzip(WithMagicNumber, Archive):
    ext = ".bzip"
    magic_number = "425a"
    iana_mime = "application/bzip"


class BaseGzip(WithMagicNumber, Archive):
    ext = ".gz"
    magic_number = "1F8B08"
    iana_mime = "application/gzip"


class BaseTar(WithMagicNumber, Archive):
    ext = ".tar"
    magic_number = "7573746172"
    magic_number_offset = 257
    iana_mime = "application/x-tar"


class BaseTarGzip(WithMagicNumber, Archive):
    # FIXME: Should capture the relationship to Gzip and Tar somehow, but a bit
    # tricky to get it right
    ext = ".tar.gz"
    magic_number = "1F8B08"
    alternate_exts = (".tgz",)
    iana_mime = "application/x-tar+gzip"


class Zip(WithArchiveClassifiers, BaseZip):
    pass


class Bzip(WithArchiveClassifiers, BaseBzip):
    pass


class Gzip(WithArchiveClassifiers, BaseGzip):
    pass


class Tar(WithArchiveClassifiers, BaseTar):
    pass


class TarGzip(WithArchiveClassifiers, BaseTarGzip):
    pass



import sys
import importlib
import typing as ty
import inspect
from itertools import zip_longest
import functools
import urllib.error
import fileformats.core
from fileformats.core.typing import TypeAlias
from .datatype import DataType
from .converter_helpers import Converter, SubtypeVar
from .exceptions import FormatConversionError, FileFormatsExtrasError
from .utils import import_extras_module, check_package_exists_on_pypi, add_exc_note

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self

if ty.TYPE_CHECKING:
    from pydra.engine.specs import TaskDef

T = ty.TypeVar("T")
ExtraImplementation = ty.TypeVar("ExtraImplementation", bound=ty.Callable[..., ty.Any])
ExtraMethod = ty.TypeVar("ExtraMethod", bound=ty.Callable[..., ty.Any])


def extra(method: ExtraMethod) -> "ExtraMethod":
    """A decorator which uses singledispatch to facilitate the registering of
    "extra" functionality in external packages (e.g. "fileformats-extras")"""

    dispatch_method: ty.Callable[..., ty.Any] = functools.singledispatch(method)

    @functools.wraps(method)
    def decorated(obj: DataType, *args: ty.Any, **kwargs: ty.Any) -> ty.Any:
        cls = type(obj)
        extras = []
        for tp in cls.referenced_types():  # type: ignore[attr-defined]
            extras.append(import_extras_module(tp))
        try:
            return dispatch_method(obj, *args, **kwargs)
        except NotImplementedError:
            msg = f"No implementation for {method.__name__!r} extra for {cls.__name__} types"
            for xtra in extras:
                if not xtra.imported:
                    try:
                        if xtra.pypi and check_package_exists_on_pypi(xtra.pypi):
                            msg += (
                                f'. An "extras" package exists on PyPI ({xtra.pypi}), '
                                "which may contain an implementation, try installing it "
                                f"(e.g. 'pip install {xtra.pypi}') and check again"
                            )
                    except urllib.error.URLError:
                        msg += (
                            '. Was not able to check whether an "extras" package '
                            f"({xtra.pypi}) exists on PyPI or not"
                        )
            raise FileFormatsExtrasError(msg)

    # Store single dispatch method on the decorated function so we can register
    # implementations to it later
    decorated._dispatch = dispatch_method  # type: ignore[attr-defined]
    return decorated  # type: ignore[return-value]


def extra_implementation(
    method: ExtraMethod,
) -> ty.Callable[[ExtraImplementation], ExtraImplementation]:
    """A decorator which uses singledispatch to facilitate the registering of
    "extra" functionality in external packages (e.g. "fileformats-extras")"""
    try:
        dispatch_method = method._dispatch  # type: ignore[attr-defined]
    except AttributeError:
        raise ValueError(
            f"{method} has not been defined as an extra method, so cannot register "
            "an implementation"
        )

    def decorator(implementation: ExtraImplementation) -> ExtraImplementation:
        msig = inspect.signature(method)
        fsig = inspect.signature(implementation)
        msig_args = list(msig.parameters.values())[1:]
        fsig_args = list(fsig.parameters.values())[1:]
        dispatched_type = list(fsig.parameters.values())[0].annotation
        differences = []

        def type_match(mtype: ty.Union[str, type], ftype: ty.Union[str, type]) -> bool:
            """Check if the types match between the method and function annotations,
            allowing for string annotations and `Self`

            Parameters
            ----------
            mtype : Union[str, type]
                the type of the method argument
            ftype : Union[str, type]
                the type of the function argument
            """
            return (
                mtype is ty.Any  # type: ignore[comparison-overlap]
                or mtype == ftype
                or inspect.isclass(ftype)
                and (
                    (inspect.isclass(mtype) and issubclass(ftype, mtype))
                    or (mtype is Self and issubclass(ftype, dispatched_type))  # type: ignore[comparison-overlap, arg-type]
                )
            )

        mhas_kwargs = msig_args and msig_args[-1].kind == inspect.Parameter.VAR_KEYWORD
        fhas_kwargs = fsig_args and fsig_args[-1].kind == inspect.Parameter.VAR_KEYWORD
        if mhas_kwargs:
            mkwargs = msig_args.pop()
            if fhas_kwargs:
                fkwargs = fsig_args.pop()
                mkwargs_type = mkwargs.annotation
                fkwargs_type = fkwargs.annotation
                if not type_match(mkwargs_type, fkwargs_type):
                    differences.append(
                        f"Type of keyword args: {mkwargs_type!r} vs {fkwargs_type!r}"
                    )
                    if isinstance(mkwargs_type, str) and not isinstance(
                        fkwargs_type, str
                    ):
                        differences.append(
                            "Note that the type of keyword args is annotated using a "
                            "string so the implementing method also needs to be a "
                            f'string, i.e. "{mkwargs_type}" instead of {fkwargs_type}'
                        )
            else:
                differences.append("variable keywords vs non-variable keywords")
        elif fhas_kwargs:
            differences.append("non-variable keywords vs variable keywords")
            fsig_args.pop()

        for i, (mparam, fparam) in enumerate(zip_longest(msig_args, fsig_args)):
            if mparam is None:
                if not mhas_kwargs:
                    differences.append(
                        f"found additional argument, {fparam.name!r}, at position {i}"
                    )
                continue
            if fparam is None:
                if mparam.default is mparam.empty:
                    differences.append(
                        f"override missing required argument {mparam.name!r}"
                    )
                continue
            mname = mparam.name
            fname = fparam.name
            mtype = mparam.annotation
            ftype = fparam.annotation
            if mname != fname:
                differences.append(
                    f"name of parameter at position {i}: {mname!r} vs {fname!r}"
                )
            elif not type_match(mtype, ftype):
                differences.append(f"Type of {mname!r} arg: {mtype!r} vs {ftype!r}")
                if isinstance(mtype, str) and not isinstance(ftype, str):
                    differences.append(
                        f"Note that the type of {mname!r} is annotated using a string so the "
                        "implementing method also needs to be a string, i.e. "
                        f'"{ftype}" instead of {ftype}'
                    )
        if not type_match(msig.return_annotation, fsig.return_annotation):
            differences.append(
                f"return type: {msig.return_annotation!r} vs {fsig.return_annotation!r}"
            )
            if isinstance(msig.return_annotation, str) and not isinstance(
                fsig.return_annotation, str
            ):
                differences.append(
                    "Note that the return type of is annotated using a string so the "
                    "implementing method also needs to be a string, i.e. "
                    f'"{fsig.return_annotation}" instead of {fsig.return_annotation}'
                )

        if differences:
            raise TypeError(
                f"Arguments differ between the signature of the extras hook method "
                f"{method} and the implementing function {implementation}:\n"
                + "\n".join(differences)
            )
        dispatch_method.register(implementation)
        return implementation

    return decorator


WrappedTask = ty.TypeVar("WrappedTask", bound=ty.Callable[..., ty.Any])
FormatType: TypeAlias = ty.Union[ty.Type["fileformats.core.FileSet"], SubtypeVar, None]


def converter(
    task_type: ty.Optional[ty.Type["TaskDef[T]"]] = None,
    source_format: FormatType = None,
    target_format: FormatType = None,
    in_file: str = "in_file",
    out_file: str = "out_file",
    **converter_kwargs: ty.Any,
) -> ty.Union[
    ty.Type["TaskDef[T]"], ty.Callable[[ty.Type["TaskDef[T]"]], ty.Type["TaskDef[T]"]]
]:
    """Decorator that registers a task as a converter between a source and target format
    pair

    Parameters
    ----------
    task_type : pydra.engine.core.TaskBase, optional
        the Pydra task to register as a converter, if provided the decorator is assumed
        to wrap the task spec directly, otherwise a wrapping decorator is returned instead
    source_format: type, optional
        file-format type to convert from, detected from type of "in_file" by default
    target_format: type, optional
        file-format type to convert to, detected from type of "out_file" by default None
    in_file : str, optional
        name of the input to be considered the "in_file", by default "in_file"
    out_file : _type_, optional
        name of the output to be considered the "out_file",
        by default "out_file" and "out" and will be tried in that order
    **converter_kwargs
        keyword arguments passed on to the converter task
    """

    try:
        import attrs
    except ImportError as e:
        add_exc_note(
            e,
            "To use the 'converter' decorator you need to have the 'attrs' package "
            "installed, this should be installed with Pydra by default",
        )
        raise e
    if "out_file_" in converter_kwargs:
        converter_kwargs["out_file"] = converter_kwargs.pop("out_file_")
    if "in_file_" in converter_kwargs:
        converter_kwargs["in_file"] = converter_kwargs.pop("in_file_")

    def decorator(
        task_type: ty.Type["TaskDef[T]"],
    ) -> ty.Type["TaskDef[T]"]:
        out_file_local = out_file
        source: FormatType
        target: FormatType
        if source_format is None:
            inputs_dict = attrs.fields_dict(task_type)
            source = inputs_dict[in_file].type
        else:
            source = source_format
        if target_format is None:
            outputs_dict = attrs.fields_dict(task_type.Outputs)
            try:
                target = outputs_dict[out_file].type
            except KeyError:
                # If there isn't an 'out_file' field but there is only one output field
                # with the default name of 'out' use that instead
                if out_file == "out_file" and list(outputs_dict.keys()) == ["out"]:
                    out_file_local = "out"
                    target = outputs_dict["out"].type
                else:
                    raise
        else:
            target = target_format
        # Handle string annotations
        if isinstance(source, str) or isinstance(target, str):
            module_dict = importlib.import_module(task_type.__module__).__dict__
            if isinstance(source, str):
                source = eval(source, module_dict)
            if isinstance(target, str):
                target = eval(target, module_dict)
        assert inspect.isclass(source) and inspect.isclass(target)
        if not issubclass(target, fileformats.core.FileSet):
            raise FormatConversionError(
                f"Target file format '{target.__name__}' is not of subtype of "
                "fileformats.core.FileSet"
            )

        target.register_converter(
            source_format=source,
            converter=Converter(
                task_type(**converter_kwargs), in_file=in_file, out_file=out_file_local
            ),
        )
        return task_type

    return decorator if task_type is None else decorator(task_type)



from fileformats.core import __version__
from .basic import Foo, Bar, Baz, Qux
from .headers import (
    YFile,
    ZFile,
    Xyz,
    MyFormat,
    MyFormatGz,
    MyFormatGzX,
    MyFormatX,
    MyBinaryFormat,
    MyHeader,
    MyBinaryFormatX,
    MyOtherBinaryFormatX,
    YourFormat,
    SeparateHeader,
    ImageWithHeader,
    EncodedText,
)
from .classifiers.file import A, B, C, D, E, F, G, H, J, K, L, M, N, P, Q, R, TestField
from .classifiers.generic import Classified, U, V, W, X, Y, Z
from .magic import Magic, MagicVersion

__all__ = [
    "__version__",
    "Foo",
    "Bar",
    "Baz",
    "Qux",
    "YFile",
    "ZFile",
    "Xyz",
    "MyFormat",
    "MyFormatGz",
    "MyFormatGzX",
    "MyFormatX",
    "MyBinaryFormat",
    "MyHeader",
    "MyBinaryFormatX",
    "MyOtherBinaryFormatX",
    "YourFormat",
    "SeparateHeader",
    "ImageWithHeader",
    "EncodedText",
    "A",
    "B",
    "C",
    "D",
    "E",
    "F",
    "G",
    "H",
    "J",
    "K",
    "L",
    "M",
    "N",
    "P",
    "Q",
    "R",
    "TestField",
    "Classified",
    "U",
    "V",
    "W",
    "X",
    "Y",
    "Z",
    "Magic",
    "MagicVersion",
]



from .core import __version__
from fileformats.core.mixin import WithMagicNumber
from fileformats.generic import UnicodeFile, BinaryFile, File


class Text(File):
    """Base class for text files"""


# General formats
class Plain(Text, UnicodeFile):
    iana_mime = "text/plain"


class TextFile(Plain, UnicodeFile):
    ext = ".txt"


class Csv(Text, UnicodeFile):
    ext = ".csv"
    iana_mime = "text/csv"


class Tsv(Text, UnicodeFile):
    ext = ".tsv"
    iana_mime = "text/tab-separated-values"


class Html(Text, UnicodeFile):
    ext = ".html"
    alternate_exts = (".htm",)
    iana_mime = "text/html"


class Markdown(Text, UnicodeFile):
    iana_mime = "text/markdown"
    ext = ".md"
    alternate_exts = (".markdown",)


class RestructedText(Text, UnicodeFile):
    ext = ".rst"
    iana_mime = "text/x-rst"


class _1d_interleaved_parityfec(Text, UnicodeFile):
    """"""

    iana_mime = "text/1d-interleaved-parityfec"
    ext = None


class CacheManifest(WithMagicNumber, Text, BinaryFile):
    """"""

    iana_mime = "text/cache-manifest"
    ext = ".appcache"
    alternate_exts = ('"manifest"',)
    magic_number = b"CACHE MANIFEST"


class Calendar(Text, UnicodeFile):
    """This media type is designed
    for widespread use by Internet calendaring and scheduling
    applications.  In addition, applications in the workflow and
    document management area might find this content-type applicable.
    The iTIP [2446bis], iMIP [2447bis], and CalDAV [RFC4791] Internet
    protocols directly use this media type also."""

    iana_mime = "text/calendar"
    ext = ".ics"
    alternate_exts = (".ifb",)


class Cql(Text, UnicodeFile):
    """"""

    iana_mime = "text/cql"
    ext = ".CQL"


class CqlExpression(Text, UnicodeFile):
    """"""

    iana_mime = "text/cql-expression"
    ext = None


class CqlIdentifier(Text, UnicodeFile):
    """"""

    iana_mime = "text/cql-identifier"
    ext = None


class Css(Text, UnicodeFile):
    """"""

    iana_mime = "text/css"
    ext = ".css"


class CsvSchema(Text, UnicodeFile):
    """"""

    iana_mime = "text/csv-schema"
    ext = None


class Dns(Text, UnicodeFile):
    """DNS-related software, including software storing and using certificates stored in DNS."""

    iana_mime = "text/dns"
    ext = ".soa"
    alternate_exts = (".zone", None)


class Encaprtp(Text, UnicodeFile):
    """"""

    iana_mime = "text/encaprtp"
    ext = None


class Fhirpath(Text, UnicodeFile):
    """"""

    iana_mime = "text/fhirpath"
    ext = None


class Flexfec(Text, UnicodeFile):
    """Multimedia applications that want to improve resiliency against packet loss by sending redundant data in addition to the source media."""

    iana_mime = "text/flexfec"
    ext = None


class Fwdred(Text, UnicodeFile):
    """It is expected that real-time audio/video, text streaming, and
    conferencing tools/applications that want protection against
    losses of a large number of consecutive frames will be interested
    in using this type."""

    iana_mime = "text/fwdred"
    ext = None


class Gff3(Text, UnicodeFile):
    """

    TODO: Although no byte sequences can be counted on
    to always be present, GFF3 data in ASCII-compatible character
    sets (including UTF-8) often begin with hexadecimal 23 23 67 66
    66 2d 76 65 72 73 69 6f 6e 20 33 ("##gff-version 3")."""

    iana_mime = "text/gff3"
    ext = ".gff3"


class GrammarRefList(Text, UnicodeFile):
    """MRCPv2 clients and servers"""

    iana_mime = "text/grammar-ref-list"
    ext = None


class Hl7v2(Text, UnicodeFile):
    """"""

    iana_mime = "text/hl7v2"
    ext = None


class Javascript(Text, UnicodeFile):
    """Script interpreters as
    discussed in RFC 9239."""

    iana_mime = "text/javascript"
    ext = ".js"
    alternate_exts = (".mjs",)


class JcrCnd(Text, UnicodeFile):
    """Used by content repositories that implement the Content Repository For Java API (JCR)"""

    iana_mime = "text/jcr-cnd"
    ext = ".cnd"


class Mizar(Text, UnicodeFile):
    """"""

    iana_mime = "text/mizar"
    ext = ".miz"


class N3(Text, UnicodeFile):
    """

    TODO: Notation3 documents may have the strings '@prefix' or '@base' (case dependent) near the
    beginning of the document."""

    iana_mime = "text/n3"
    ext = ".n3"


class Parameters(Text, UnicodeFile):
    """Applications that use RTSP
    and have additional parameters they like to read and set using the
    RTSP GET_PARAMETER and SET_PARAMETER methods."""

    iana_mime = "text/parameters"
    ext = None


class Parityfec(Text, UnicodeFile):
    """"""

    iana_mime = "text/parityfec"
    ext = None


class ProvenanceNotation(Text, UnicodeFile):
    """

    TODO: PROV-N documents may have the strings 'document' near the beginning of the document.
    """

    iana_mime = "text/provenance-notation"
    ext = ".provn"


class Prs_Fallenstein_Rst(Text, UnicodeFile):
    """"""

    iana_mime = "text/prs.fallenstein.rst"
    ext = ".txt"
    alternate_exts = (".rst",)


class Prs_Lines_Tag(Text, UnicodeFile):
    """

    TODO: All tag data streams start Tag-xxx-version - where xxx is the particular tag type.
    """

    iana_mime = "text/prs.lines.tag"
    ext = ".tag"
    alternate_exts = (None,)


class Prs_Prop_Logic(Text, UnicodeFile):
    """"""

    iana_mime = "text/prs.prop.logic"
    ext = ".txt"


class Raptorfec(Text, UnicodeFile):
    """Real-time multimedia applications like video streaming, audio streaming, and video conferencing."""

    iana_mime = "text/raptorfec"
    ext = None


class Red(Text, UnicodeFile):
    """"""

    iana_mime = "text/RED"
    ext = None


class Rfc822_headers(Text, UnicodeFile):
    """"""

    iana_mime = "text/rfc822-headers"
    ext = None


class RichText(Text, UnicodeFile):
    """"""

    iana_mime = "text/rtf"
    ext = ".rtf"


class RtpEncAescm128(Text, UnicodeFile):
    """"""

    iana_mime = "text/rtp-enc-aescm128"
    ext = None


class Rtploopback(Text, UnicodeFile):
    """"""

    iana_mime = "text/rtploopback"
    ext = None


class Rtx(Text, UnicodeFile):
    """"""

    iana_mime = "text/rtx"
    ext = None


class Sgml(Text, UnicodeFile):
    """"""

    iana_mime = "text/SGML"
    ext = None


class Shaclc(Text, UnicodeFile):
    """An implementations of SHACLC
    is part of the TopQuadrant SHACL API

    TODO: SHACLC documents will likely have the words
    PREFIX or BASE (case sensitive) near the beginning of the
    document. However, the same words may appear in Turtle and
    SPARQL documents."""

    iana_mime = "text/shaclc"
    ext = ".shaclc"
    alternate_exts = (".shc",)


class Shex(Text, UnicodeFile):
    """"""

    iana_mime = "text/shex"
    ext = ".shex"


class Spdx(Text, UnicodeFile):
    """The "charset" parameter is not used for the
    defined subtype because the charset information is transported
    inside the payload."""

    iana_mime = "text/spdx"
    ext = ".spdx"


class Strings(Text, UnicodeFile):
    """"""

    iana_mime = "text/strings"
    ext = None


class T140(Text, UnicodeFile):
    """This type is only defined for transfer via
    RTP."""

    iana_mime = "text/t140"
    ext = None


class Troff(Text, UnicodeFile):
    """"""

    iana_mime = "text/troff"
    ext = None


class Turtle(Text, UnicodeFile):
    """

    TODO: Turtle documents may have the strings '@prefix' or '@base' (case dependent) near the
    beginning of the document."""

    iana_mime = "text/turtle"
    ext = ".ttl"


class Ulpfec(Text, UnicodeFile):
    """Multimedia applications that seek to improve resiliency to loss by sending additional data with the media stream."""

    iana_mime = "text/ulpfec"
    ext = None


class UriList(Text, UnicodeFile):
    """"""

    iana_mime = "text/uri-list"
    ext = ".uris"
    alternate_exts = (".uri",)


class Vcard(Text, UnicodeFile):
    """They are numerous, diverse,
    and include mail user agents, instant messaging clients, address
    book applications, directory servers, and customer relationship
    management software."""

    iana_mime = "text/vcard"
    ext = ".vcf"
    alternate_exts = (".vcard",)


class Vtt(Text, UnicodeFile):
    """Web browsers and other video
    players.

    TODO: WebVTT files all begin with one of the following byte sequences (where "EOF" means the end of the file):
    EF BB BF 57 45 42 56 54 54 0A EF BB BF 57 45 42 56 54 54 0D EF BB BF 57 45 42 56 54 54 20 EF BB BF 57 45 42 56 54 54 09 EF BB BF 57 45 42 56 54 54 EOF 57 45 42 56 54 54 0A 57 45 42 56 54 54 0D 57 45 42 56 54 54 20 57 45 42 56 54 54 09 57 45 42 56 54 54 EOF
    (An optional UTF-8 BOM, the ASCII string "WEBVTT", and finally a space, tab, line break, or the end of the file.)
    """

    iana_mime = "text/vtt"
    ext = ".vtt"


class Wgsl(Text, UnicodeFile):
    """"""

    iana_mime = "text/wgsl"
    ext = ".wgsl"


class XmlExternalParsedEntity(Text, UnicodeFile):
    """"""

    iana_mime = "text/xml-external-parsed-entity"
    ext = ".ent"
    alternate_exts = (None,)


from fileformats.application import (  # noqa
    Json,
    Xml,
    Yaml,
)  # These are sometimes/historically considered part of the text registry so we import them here


__all__ = [
    "__version__",
    "Text",
    "Plain",
    "TextFile",
    "Csv",
    "Tsv",
    "Html",
    "Markdown",
    "RestructedText",
    "_1d_interleaved_parityfec",
    "CacheManifest",
    "Calendar",
    "Cql",
    "CqlExpression",
    "CqlIdentifier",
    "Css",
    "CsvSchema",
    "Dns",
    "Encaprtp",
    "Fhirpath",
    "Flexfec",
    "Fwdred",
    "Gff3",
    "GrammarRefList",
    "Hl7v2",
    "Javascript",
    "JcrCnd",
    "Mizar",
    "N3",
    "Parameters",
    "Parityfec",
    "ProvenanceNotation",
    "Prs_Fallenstein_Rst",
    "Prs_Lines_Tag",
    "Prs_Prop_Logic",
    "Raptorfec",
    "Red",
    "Rfc822_headers",
    "RichText",
    "RtpEncAescm128",
    "Rtploopback",
    "Rtx",
    "Sgml",
    "Shaclc",
    "Shex",
    "Spdx",
    "Strings",
    "T140",
    "Troff",
    "Turtle",
    "Ulpfec",
    "UriList",
    "Vcard",
    "Vtt",
    "Wgsl",
    "XmlExternalParsedEntity",
]



import typing as ty
from pathlib import Path
import time
from threading import RLock
import fileformats.core
from .fs_mount_identifier import FsMountIdentifier


PropReturn = ty.TypeVar("PropReturn")


__all__ = ["mtime_cached_property", "classproperty"]


class mtime_cached_property:
    """A property that is cached until the mtimes of the files in the fileset are changed"""

    def __init__(self, func: ty.Callable[..., ty.Any]):
        self.func = func
        self.__doc__ = func.__doc__
        self.lock = RLock()
        self._cache_name = f"_{func.__name__}_mtime_cache"

    def __get__(
        self,
        instance: ty.Optional["fileformats.core.FileSet"],
        owner: ty.Optional[ty.Type["fileformats.core.FileSet"]] = None,
    ) -> ty.Any:
        if instance is None:  # if accessing property from class not instance
            return self
        assert isinstance(instance, fileformats.core.FileSet), (
            "Cannot use mtime_cached_property instance with "
            f"{type(instance).__name__!r} object, only FileSet objects."
        )
        try:
            mtimes, value = instance.__dict__[self._cache_name]
        except KeyError:
            pass
        else:
            if (
                instance.mtimes == mtimes
                and enough_time_has_elapsed_given_mtime_resolution(mtimes)
            ):
                return value
            else:
                del instance.__dict__[self._cache_name]
        with self.lock:
            # check if another thread filled cache while we awaited lock
            try:
                mtimes, value = instance.__dict__[self._cache_name]
            except KeyError:
                pass
            else:
                if (
                    instance.mtimes == mtimes
                    and enough_time_has_elapsed_given_mtime_resolution(mtimes)
                ):
                    return value
            value = self.func(instance)
            instance.__dict__[self._cache_name] = (instance.mtimes, value)
        return value


# def classproperty(meth: ty.Callable[..., PropReturn]) -> PropReturn:
#     """Access a @classmethod like a @property."""
#     # mypy doesn't understand class properties yet: https://github.com/python/mypy/issues/2563
#     return classmethod(property(meth))  # type: ignore


def validated_property(meth: ty.Callable[..., PropReturn]) -> PropReturn:
    """A property that is checked during validation of a FileSet"""
    prop = property(meth)
    prop.fget.__annotations__[VALIDATED_PROPERTY_FLAG] = True
    return prop  # type: ignore


class classproperty(object):  # type: ignore[no-redef]  # noqa
    def __init__(self, f: ty.Callable[[ty.Any], ty.Any]):
        self.f = f

    def __get__(self, obj: ty.Any, owner: ty.Any) -> ty.Any:
        return self.f(owner)


def enough_time_has_elapsed_given_mtime_resolution(
    mtimes: ty.Iterable[ty.Tuple[Path, ty.Union[int, float]]],
    current_time: ty.Optional[int] = None,
) -> bool:
    """Determines whether enough time has elapsed since the the last of the cached mtimes
    to be sure that changes in mtimes will be detected given the resolution of the mtimes
    on the file system. For example, on systems with a mtime resolution of a second,
    a change in mtime may not be detected if the cache is re-read within a second and
    the file is modified in the intervening period (probably only likely during tests).
    So this function guesses the resolution of the mtimes by the
    minimum number of trailing zeros in the mtimes and then checks if enough time has
    passed to be sure that any changes in mtimes will be detected.

    This may result in false negatives for systems with low mtime resolutions, but
    this will just result in (presumably minor) performance hit via unnecessary cache
    invalidations.

    Parameters
    ----------
    mtimes : Iterable[tuple[Path, int]]
        the path-mtime pairs in nanoseconds to guess the resolution of

    Returns
    -------
    bool
        whether enough time has elapsed since the lagiven the guessed resolution of the mtimes
    """
    if current_time is None:
        current_time = time.time_ns()
    for fspath, mtime in mtimes:
        mtime_resolution = FsMountIdentifier.get_mtime_resolution(fspath)
        if current_time - mtime <= mtime_resolution:
            return False
    return True


VALIDATED_PROPERTY_FLAG = "_validated_property"



from fileformats.core import __version__, FileSet
from .fsobject import FsObject
from .file import File, BinaryFile, UnicodeFile
from .directory import Directory, TypedDirectory, DirectoryOf
from .set import TypedSet, SetOf
from . import generate_sample_data  # noqa: F401


__all__ = [
    "__version__",
    "FileSet",
    "FsObject",
    "File",
    "Directory",
    "TypedDirectory",
    "DirectoryOf",
    "TypedSet",
    "SetOf",
    "BinaryFile",
    "UnicodeFile",
]



from pathlib import Path
import re
import typing as ty
import logging
from .datatype import DataType
import fileformats.core
from .utils import get_optional_type
from .decorators import validated_property, classproperty
from .identification import to_mime_format_name
from .converter_helpers import SubtypeVar, Converter
from .classifier import Classifier
from .exceptions import (
    FormatMismatchError,
    FormatRecognitionError,
    FormatDefinitionError,
)


logger = logging.getLogger("fileformats")

T = ty.TypeVar("T")


class WithMagicNumber:
    """Mixin class for Files with magic numbers at the start of their
    contents.

    Class Attrs
    -----------
    magic_number : str or bytes
        the magic number/string to search for at the start of the file. If a unicode
        string then it is interpreted as the byte code in hex, if a bytes object, then
        it is treated as the byte string directly.
    binary : bool
        if the file-format is a binary type then this flag needs to be set in order to
        read the contents properly
    magic_number_offset : int, optional
        the offset in bytes from the start of the file that the magic number is stored
    """

    magic_number_offset = 0
    binary: bool
    magic_number: ty.Union[str, bytes]

    @validated_property
    def _check_magic_number(self) -> None:
        if getattr(self, "binary", True) and isinstance(self.magic_number, str):
            try:
                magic_bytes = bytes.fromhex(self.magic_number)
            except ValueError:
                raise FormatDefinitionError(
                    f"Magic number of file {type(self)} is not a valid hex string"
                )
        else:
            assert isinstance(self.magic_number, bytes)
            magic_bytes = self.magic_number
        read_magic_number = self.read_contents(  # type: ignore[attr-defined]
            len(magic_bytes), offset=self.magic_number_offset
        )
        if read_magic_number != magic_bytes:
            read_magic: ty.Union[str, bytes]
            ref_magic: ty.Union[str, bytes]
            if getattr(self, "binary", True) and isinstance(self.magic_number, str):
                read_magic = '"' + bytes.hex(read_magic_number) + '"'
                ref_magic = '"' + self.magic_number + '"'
            else:
                read_magic = read_magic_number
                assert isinstance(self.magic_number, bytes)
                ref_magic = self.magic_number
            raise FormatMismatchError(
                f"Magic number of file {read_magic!r} doesn't match expected "
                f"{ref_magic!r}"
            )


class WithMagicVersion:
    """Mixin class for Files with version numbers embedded within "magic numbers"
    the start of their contents.

    Class Attrs
    -----------
    magic_pattern : bytes
        the magic number/string to search for at the start of the file
    magic_pattern_offset : int, optional
        the offset in bytes from the start of the file that the magic pattern is read from
    magic_pattern_maxlength : int, optional
        the maximum length of the pattern, i.e. the length of the byte-string that will
        be read from the file, by default it will be the length of the magic_pattern
        string (which will probably be longer than the string it matches due to the
        special characters in the regex)
    """

    binary: bool
    magic_pattern: bytes
    magic_pattern_offset = 0
    magic_pattern_maxlength: ty.Optional[int] = None

    @validated_property
    def version(self) -> ty.Union[str, ty.Tuple[str, ...]]:
        read_length = (
            self.magic_pattern_maxlength
            if self.magic_pattern_maxlength
            else len(self.magic_pattern)
        )
        read_bytes = self.read_contents(read_length, offset=self.magic_pattern_offset)  # type: ignore[attr-defined]
        match = re.match(self.magic_pattern, read_bytes)
        if not match:
            raise FormatMismatchError(
                f"Byte-string of length {read_length} at {self.magic_pattern_offset} "
                f"({read_bytes!r}), doesn't match expected pattern, {self.magic_pattern!r}"
            )
        version: ty.Tuple[str, ...] = tuple(b.decode("utf-8") for b in match.groups())
        if not version:
            raise FormatDefinitionError(
                f"No version patterns found in magic pattern of {type(self).__name__} "
                f"class, {self.magic_pattern!r}"
            )
        if len(version) == 1:
            return version[0]
        return version


class WithAdjacentFiles:
    """
    If only the main fspath is provided to the __init__ of the class, this mixin
    automatically includes any "adjacent files", i.e. any files with the same stem but
    different extensions

    Note that WithAdjacentFiles must come before the primary type in the method-resolution
    order of the class so it can override the '_additional_paths' method in

        class MyFileFormatWithSeparateHeader(WithSeparateHeader, MyFileFormat):

            header_type = MyHeaderType
    """

    fspaths: ty.FrozenSet[Path]

    def _additional_fspaths(self) -> None:
        if len(self.fspaths) == 1:
            self.fspaths |= self.get_adjacent_files()
            trim = True
        else:
            trim = False
        if trim:
            self.trim_paths()  # type: ignore[attr-defined]

    def get_adjacent_files(self) -> ty.Set[Path]:
        stem = self.stem  # type: ignore[attr-defined]
        adjacents = set()
        for sibling in self.fspath.parent.iterdir():  # type: ignore[attr-defined]
            if (
                sibling != self.fspath  # type: ignore[attr-defined]
                and sibling.is_file()
                and sibling.name.startswith(stem + ".")
            ):
                adjacents.add(sibling)
        return adjacents


class WithSeparateHeader(WithAdjacentFiles):
    """Mixin class for Files with metadata stored in separate header files (typically
    with the same file stem but differing extension)

    Note that WithSeparateHeader must come before the primary type in the method-resolution
    order of the class

        class MyFileFormatWithSeparateHeader(WithSeparateHeader, MyFileFormat):

            header_type = MyHeaderType

    Class Attrs
    -----------
    header_type : type
        the file-format of the header file
    """

    header_type: ty.Type["fileformats.core.FileSet"]

    @classproperty  # type: ignore[arg-type]
    def nested_types(cls) -> ty.Tuple[ty.Type[Classifier], ...]:
        return (cls.header_type,)

    @validated_property
    def header(self) -> "fileformats.core.FileSet":
        return self.header_type(self.select_by_ext(self.header_type))  # type: ignore[attr-defined]

    def read_metadata(self, **kwargs: ty.Any) -> ty.Mapping[str, ty.Any]:
        header: ty.Dict[str, ty.Any] = self.header.load()
        return header


class WithSideCars(WithAdjacentFiles):
    """Mixin class for Files with a "side-car" file that augments the inline metadata
    (typically with the same file stem but differing extension).

    Note that WithSideCars must come before the primary type in the method-resolution
    order of the class methods, e.g.

        class MyFileFormatWithSideCars(WithSideCars, MyFileFormat):

            primary_type = MyFileFormat
            side_car_types = (MySideCarType,)

    Class Attrs
    -----------
    primary_type : type
        the file-format of the primary file (used to read the inline metadata), can be
        the base class that implements 'read_metadata'
    side_car_types : tuple[type, ...]
        the file-formats of the expected side-car files
    """

    primary_type: ty.Type["fileformats.core.FileSet"]
    side_car_types: ty.Tuple[ty.Type["fileformats.core.FileSet"], ...]

    @validated_property
    def side_cars(self) -> ty.Tuple["fileformats.core.FileSet", ...]:
        return tuple(tp(self.select_by_ext(tp)) for tp in self.side_car_types)  # type: ignore[attr-defined]

    def read_metadata(self, **kwargs: ty.Any) -> ty.Mapping[str, ty.Any]:
        metadata: ty.Dict[str, ty.Any] = dict(self.primary_type.read_metadata(self, **kwargs))  # type: ignore[arg-type]
        for side_car in self.side_cars:
            try:
                side_car_metadata: ty.Dict[str, ty.Any] = side_car.load()
            except AttributeError:
                continue
            if not isinstance(side_car_metadata, dict):
                raise TypeError(
                    f"`load` method of side-car type {type(side_car)} must return a "
                    f"dictionary, not {type(side_car_metadata)!r}"
                )
            side_car_class_name: str = to_mime_format_name(type(side_car).__name__)
            metadata[side_car_class_name] = side_car_metadata
        return metadata

    @classproperty  # type: ignore[arg-type]
    def nested_types(cls) -> ty.Tuple[ty.Type[Classifier], ...]:
        return cls.side_car_types


class WithClassifiers:
    """Mixin class for adding the ability to qualify the format class to designate the
    type of information stored within the format, e.g. ``DirectoryOf[Png, Gif]`` for a
    directory containing PNG and GIF files, ``Zip[DataFile]`` for a zipped data file,
    ``Array[Integer]`` for an array containing integers, or DicomDir[T1w, Brain] for a
    T1-weighted MRI scan of the brain in DICOM format.

        class MyFormatWithClassifiers(WithClassifiers, BinaryFile):

            ext = ".myf


        def my_func(file: MyFormatWithClassifiers[Integer]):
            ...

    A unique class will be returned (i.e. multiple calls with the same arguments will
    return the same class)

    Class Attrs
    -----------
    classifiers_attr_name : str, optional
        an attribute name to store the classifiers at within the classified class. This
        should be used if you need to reference the ``classifiers`` attribute directly
        in any validation/other methods (i.e. in most cases), to handle the case of
        diamond inheritance between two classes that can be classified.
        A default value should also be set in the unclassified base for this attribute,
        which is either ``()`` if ``multiple_classifiers`` is true and ``None`` otherwise
    <classifiers-attr-name> : tuple[type, ...] or None, optional
        pass a default value to the attribute referenced by 'classifiers_attr_name'
    allowed_classifiers : tuple[type,...], optional
        the allowable types (+ subclasses) for the content types. If None all types
        are allowed
    genericly_classified : bool, optional
        whether the class can be classified by classifiers in any namespace (true) or just the
        namespace it belongs to (false). If true, then the namespace of the genericly
        classified class is omitted from the "mime-like" string. Note that the
        class' name therefore needs to be globally unique amongst all other genericly
        classified classes and so it should be used sparingly, i.e., highly generic
        formats that are unambiguous across all namespaces, such as "directory", "zip",
        "gzip", "json", "yaml", etc...
    """

    # classifiers set in the current class
    classifiers: ty.Tuple[ty.Type[DataType], ...] = ()
    _classified_subtypes: ty.Dict[
        ty.Tuple[ty.Type[Classifier], ...], ty.Type[DataType]
    ] = {}
    # dict of previously created classified subtypes. If an existing class with matching
    # classifiers has been created it is returned instead of creating a new type. This
    # ensures that ``assert MyFormat[Qualifier] is MyFormat[Qualifier]``

    # Default values for class attrs
    multiple_classifiers = True
    allowed_classifiers: ty.Optional[ty.Tuple[ty.Type[Classifier], ...]] = None
    allow_optional_classifiers = False
    exclusive_classifiers: ty.Tuple[ty.Type[Classifier], ...] = ()
    ordered_classifiers = False
    generically_classifiable = False

    def _validate_class(self) -> ty.Union[bool, None]:
        validated: ty.Union[bool, None] = super()._validate_class()  # type: ignore
        if validated is None:
            if self.wildcard_classifiers():
                raise FormatDefinitionError(
                    f"Can instantiate {type(self)} class as it has wildcard classifiers "
                    "and therefore should only be used for converter specifications"
                )
        return validated

    @classproperty  # type: ignore[arg-type]
    def is_classified(cls) -> bool:
        return "unclassified" in cls.__dict__

    @classproperty  # type: ignore[arg-type]
    def nested_types(cls) -> ty.Tuple[ty.Type[Classifier], ...]:
        return cls.classifiers

    @classmethod
    def wildcard_classifiers(
        cls, classifiers: ty.Optional[ty.Sequence[ty.Type[Classifier]]] = None
    ) -> ty.FrozenSet[ty.Type[SubtypeVar]]:
        if classifiers is None:
            classifiers = cls.classifiers if cls.is_classified else ()
        return frozenset(
            t for t in classifiers if issubclass(get_optional_type(t), SubtypeVar)  # type: ignore[misc]
        )

    @classmethod
    def non_wildcard_classifiers(
        cls, classifiers: ty.Optional[ty.Collection[ty.Type[Classifier]]] = None
    ) -> ty.FrozenSet[ty.Type[Classifier]]:
        if classifiers is None:
            classifiers = cls.classifiers if cls.is_classified else ()
        assert classifiers is not None
        return frozenset(
            q for q in classifiers if not issubclass(get_optional_type(q), SubtypeVar)
        )

    @classmethod
    def __class_getitem__(
        cls,
        classifiers: ty.Union[ty.Collection[ty.Type[Classifier]], ty.Type[Classifier]],
    ) -> ty.Type[DataType]:
        """Set the content types for a newly created dynamically type"""
        if isinstance(classifiers, ty.Iterable):
            classifiers_tuple = tuple(classifiers)
        else:
            classifiers_tuple = (classifiers,)
        classifiers_to_check = tuple(
            get_optional_type(c, cls.allow_optional_classifiers)
            for c in classifiers_tuple
        )

        if cls.allowed_classifiers:
            not_allowed = [
                q
                for q in classifiers_to_check
                if not any(issubclass(q, t) for t in cls.allowed_classifiers)
            ]
            if not_allowed:
                raise FormatDefinitionError(
                    f"Invalid content types provided to {cls} (must be subclasses of "
                    f"{cls.allowed_classifiers}): {not_allowed}"
                )
        # Sort content types if order isn't important
        if cls.multiple_classifiers:
            if not cls.ordered_classifiers:
                # Check for duplicate classifiers in the multiple list
                if len(classifiers_to_check) > 1:
                    # Sort the classifiers into categories and ensure that there aren't more
                    # than one type for each category. Otherwise, if the classifier doesn't
                    # belong to a category, check to see that there aren't multiple sub-classes
                    # in the classifier set
                    repetitions: ty.Dict[
                        ty.Type[Classifier], ty.List[ty.Type[Classifier]]
                    ] = {
                        c: [] for c in cls.exclusive_classifiers + classifiers_to_check
                    }
                    for classifier in classifiers_to_check:
                        for exc_classifier in repetitions:
                            if issubclass(classifier, exc_classifier):
                                repetitions[exc_classifier].append(classifier)
                    repeated = [t for t in repetitions.items() if len(t[1]) > 1]
                    if repeated:
                        raise FormatDefinitionError(
                            "Cannot have more than one occurrence of a classifier "
                            f"or subclasses for {cls} class when "
                            f"{cls.__name__}.ordered_classifiers is false:\n"
                            + "\n".join(
                                f"{k!r}: " + ", ".join(repr(x) for x in v)
                                for k, v in repeated
                            )
                        )
                classifiers_tuple = tuple(
                    sorted(
                        set(classifiers_tuple),
                        key=lambda x: get_optional_type(x).__name__,
                    )
                )
        else:
            if len(classifiers_tuple) > 1:
                raise FormatDefinitionError(
                    f"Multiple classifiers not permitted for {cls} types, provided: "
                    f"({classifiers_tuple})"
                )
        # Make sure that the "classified" dictionary is present in this class not super
        # classes
        if "_classified_subtypes" not in cls.__dict__:
            cls._classified_subtypes = {}
        try:
            # Load previously created type so we can do ``assert MyType[Integer] is MyType[Integer]``
            classified = cls._classified_subtypes[classifiers_tuple]
        except KeyError:
            if not hasattr(cls, "classifiers_attr_name"):
                raise FormatDefinitionError(
                    f"{cls} needs to define the 'classifiers_attr_name' class attribute "
                    "with the name of the (different) class attribute to hold the "
                    "classified types"
                )
            if cls.classifiers_attr_name is None:
                raise FormatDefinitionError(
                    f"Inherited classifiers have been disabled in {cls} (by setting "
                    f'"classifiers_attr_name)" to None)'
                )
            try:
                classifiers_attr = getattr(cls, cls.classifiers_attr_name)
            except AttributeError:
                raise FormatDefinitionError(
                    f"Default value for classifiers attribute "
                    f"'{cls.classifiers_attr_name}' needs to be set in {cls}"
                )
            else:
                if classifiers_attr:
                    raise FormatDefinitionError(
                        f"Default value for classifiers attribute "
                        f"'{cls.classifiers_attr_name}' needs to be set in {cls}"
                    )
            class_attrs = {
                "unclassified": cls,
                "classifiers": classifiers_tuple,
            }
            class_attrs[cls.classifiers_attr_name] = (
                classifiers_tuple if cls.multiple_classifiers else classifiers_tuple[0]
            )
            classifier_names = [
                get_optional_type(t).__name__ for t in classifiers_tuple
            ]
            if not cls.ordered_classifiers:
                classifier_names.sort()
            classified = type(
                f"{'_'.join(classifier_names)}__{cls.__name__}",
                (cls,),
                class_attrs,
            )
            classified.__module__ = cls.__module__
            cls._classified_subtypes[classifiers_tuple] = classified
        return classified

    @classmethod
    def get_converter_defs(cls, source_format: type) -> ty.List[Converter]:
        """Search the registered converters to find an appropriate task and associated
        key-word args to perform the conversion between source and target formats

        Parameters
        ----------
        source_format : type(FileSet)
            the source format to convert from
        """
        from fileformats.core import FileSet

        # Try to see if a converter has been defined to the exact type
        available_converters: ty.List[Converter] = super().get_converter_defs(  # type: ignore[misc, type-arg]
            source_format
        )
        # Failing that, see if there is a generic conversion between the container type
        # the source format (or subclass of) defined with matching wildcards in the source
        # and target formats
        if not available_converters and cls.is_classified:
            converters_dict = FileSet.get_converters_dict(
                cls.unclassified  # type: ignore[attr-defined]
            )  # pylint: disable=no-member
            for template_source_format, converter in converters_dict.items():
                if converter.classifiers:  # was defined with wildcard classifiers
                    # Attempt conversion from generic type to template match
                    if issubclass(template_source_format, SubtypeVar):
                        assert tuple(
                            cls.wildcard_classifiers(converter.classifiers)
                        ) == (template_source_format,)
                        non_wildcards = cls.non_wildcard_classifiers(
                            converter.classifiers
                        )
                        to_match = tuple(set(cls.classifiers).difference(non_wildcards))
                        if len(to_match) > 1:
                            wildcard_match = False
                        else:
                            wildcard_match = issubclass(source_format, to_match[0])
                    # Attempt template to template conversion match
                    elif getattr(source_format, "is_classified", False) and issubclass(
                        source_format.unclassified, template_source_format.unclassified  # type: ignore[attr-defined]
                    ):
                        assert cls.wildcard_classifiers(
                            converter.classifiers
                        ) == cls.wildcard_classifiers(
                            template_source_format.classifiers  # type: ignore[attr-defined]
                        )
                        if cls.ordered_classifiers:
                            if len(cls.classifiers) != len(
                                converter.classifiers
                            ) or len(
                                source_format.classifiers  # type: ignore[attr-defined]
                            ) != len(
                                template_source_format.classifiers  # type: ignore[attr-defined]
                            ):
                                wildcard_match = False
                            else:
                                wildcard_map = {}
                                for actual, template in zip(
                                    source_format.classifiers,  # type: ignore[attr-defined]
                                    template_source_format.classifiers,  # type: ignore[attr-defined]
                                ):
                                    if issubclass(template, SubtypeVar):
                                        wildcard_map[template] = actual
                                wildcard_match = True
                                for actual, template in zip(
                                    cls.classifiers, converter.classifiers
                                ):
                                    if issubclass(template, SubtypeVar):
                                        try:
                                            reference = wildcard_map[template]
                                        except KeyError:
                                            wildcard_match = False
                                            break
                                        else:
                                            if not issubclass(actual, reference):
                                                wildcard_match = False
                                                break
                                    elif not issubclass(actual, template):
                                        wildcard_match = False
                                        break
                        else:
                            non_wildcards = cls.non_wildcard_classifiers(
                                converter.classifiers
                            )
                            src_non_wildcards = cls.non_wildcard_classifiers(
                                template_source_format.classifiers  # type: ignore[attr-defined]
                            )
                            if not non_wildcards.issubset(
                                set(cls.classifiers)
                            ) or not src_non_wildcards.issubset(
                                set(source_format.classifiers)  # type: ignore[attr-defined]
                            ):
                                wildcard_match = False
                            else:
                                to_match = set(cls.classifiers).difference(
                                    non_wildcards
                                )
                                from_types = set(source_format.classifiers).difference(  # type: ignore[attr-defined]
                                    src_non_wildcards
                                )
                                wildcard_match = to_match.issubset(from_types)
                    else:
                        wildcard_match = False
                    if wildcard_match:
                        available_converters.append(converter)
        return available_converters

    @classmethod
    def __subclasshook__(cls, subclass: type) -> bool:
        """Overload the behaviour of 'issubclass' so that classified classes are considered
        to be subclasses of each other if they contain a super-set of classifiers"""
        if type.__subclasscheck__(cls, subclass):
            return True
        # Check to see whether the unclassified types are equivalent
        if (
            not cls.is_classified
            or not getattr(subclass, "is_classified", False)
            or not issubclass(subclass.unclassified, cls.unclassified)  # type: ignore[attr-defined]
        ):
            return False
        if cls.ordered_classifiers:
            assert subclass.ordered_classifiers  # type: ignore[attr-defined]
            if len(subclass.classifiers) != len(cls.classifiers):  # type: ignore[attr-defined]
                is_subclass = False
            else:
                is_subclass = all(
                    issubclass(q, s)
                    for q, s in zip(subclass.classifiers, cls.classifiers)  # type: ignore[attr-defined]
                )
        else:
            assert not subclass.ordered_classifiers  # type: ignore[attr-defined]
            if set(subclass.classifiers).issuperset(cls.classifiers):  # type: ignore[attr-defined]
                is_subclass = True
            else:
                # Check for sub-classes of classifiers
                is_subclass = all(
                    any(issubclass(q, s) for q in subclass.classifiers)  # type: ignore[attr-defined]
                    for s in cls.classifiers
                )
        return is_subclass

    @classmethod
    def register_converter(
        cls,
        source_format: ty.Type["fileformats.core.FileSet"],
        converter: Converter,
    ) -> None:
        """Registers a converter task within a class attribute. Called by the @fileformats.converter
        decorator.

        Parameters
        ----------
        source_format : type
            the source format to register a converter from
        converter_spec
            a tuple consisting of a `task_spec` callable that resolves to a Pydra task
            and a dictionary of keyword arguments to be passed to the task spec at
            initialisation time

        Raises
        ------
        FormatConversionError
            if there is already a converter registered between the two types
        """
        # Ensure "converters" dict is defined in the target class and not in a superclass
        if cls.wildcard_classifiers():
            if issubclass(source_format, SubtypeVar):
                if len(cls.wildcard_classifiers()) > 1:
                    raise FormatDefinitionError(
                        "Can only have one wildcard qualifier when registering a converter "
                        f"to {cls} from a generic type, found {cls.wildcard_classifiers()}"
                    )
            elif not source_format.is_classified:  # type: ignore[attr-defined]
                raise FormatDefinitionError(
                    "Can only use wildcard classifiers when registering a converter "
                    f"from a generic type or similarly classified type, not {source_format}"
                )
            else:
                source_wildcard_classifiers = source_format.wildcard_classifiers()  # type: ignore[attr-defined]
                if cls.wildcard_classifiers() != source_wildcard_classifiers:
                    raise FormatDefinitionError(
                        f"Mismatching wildcards between source format, {source_format} "
                        f"({list(source_wildcard_classifiers)}), and target "
                        f"{cls} ({cls.wildcard_classifiers()})"
                    )
                prev_registered = [
                    f
                    for f in cls.converters  # type: ignore[attr-defined]
                    if (
                        issubclass(source_format.unclassified, f.unclassified)  # type: ignore[attr-defined]
                        and f.non_wildcard_classifiers()
                        == source_format.non_wildcard_classifiers()  # type: ignore[attr-defined]
                    )
                ]
                assert len(prev_registered) <= 1
                prev = prev_registered[0] if prev_registered else None
                if prev:
                    prev_converter = cls.converters[prev]  # type: ignore[attr-defined]
                    # task, task_kwargs = converter_spec
                    # prev_task, prev_kwargs, prev_classifiers = prev_tuple
                    if converter == prev_converter:
                        logger.warning(
                            "Ignoring duplicate registrations of the same converter %s",
                            converter.task_def,
                        )
                        return  # actually the same task but just imported twice for some reason
                    prev_unclassified = prev.unclassified
                    unclassified = cls.unclassified  # type: ignore[attr-defined]
                    raise FormatDefinitionError(
                        f"Cannot register converter from {prev_unclassified} "
                        f"to {unclassified} with non-wildcard classifiers "
                        f"{list(prev.non_wildcard_classifiers())}, {converter.task_def}, "
                        f"because there is already one registered, {prev_converter.task}"
                    )
            converters_dict = cls.unclassified.get_converters_dict()  # type: ignore[attr-defined]
            converter.classifiers = cls.classifiers
            converters_dict[source_format] = converter
        else:
            super().register_converter(source_format, converter)  # type: ignore[misc]

    @classproperty  # type: ignore[arg-type]
    def namespace(cls) -> ty.Optional[str]:
        """The "namespace" the format belongs to under the "fileformats" umbrella
        namespace"""
        namespace: ty.Optional[str]
        if cls.is_classified:
            namespaces: ty.Collection[str] = set(
                t.namespace for t in cls.classifiers if t.namespace
            )
            if not cls.generically_classifiable:
                namespaces.add(cls.unclassified.namespace)  # type: ignore[attr-defined]
            if len(namespaces) == 1:
                return next(iter(namespaces))
            else:
                # Handle subpackage namespaces and parent, e.g. medimage & medimage-fsl
                namespaces = sorted(namespaces)
                if (
                    len(namespaces) == 2
                    and namespaces[1].split("-")[0] == namespaces[0]
                ):
                    return namespaces[1]
                msg = (
                    "Cannot create reversible MIME type for because did not find a "
                    f"common namespace between all classifiers {list(cls.classifiers)}"
                )
                if not cls.generically_classifiable:
                    msg += f" and (non genericly classified) base class {cls.unclassified}"  # type: ignore[attr-defined]
                raise FormatRecognitionError(msg + f", found:\n{list(namespaces)}")
        else:
            try:
                namespace = super().namespace  # type: ignore[misc]
            except AttributeError:
                namespace = None
        return namespace

    @classproperty  # type: ignore[arg-type]
    def type_name(cls) -> str:
        """Name of type including classifiers to be used in __repr__"""
        unclassified: str
        if not cls.is_classified:
            return cls.__name__  # type: ignore[no-any-return, attr-defined]
        unclassified = cls.unclassified.__name__  # type: ignore[attr-defined]
        return (
            unclassified + "[" + ", ".join(t.type_name for t in cls.classifiers) + "]"
        )


class WithClassifier(WithClassifiers):

    multiple_classifiers = False


class WithOrderedClassifiers(WithClassifiers):

    ordered_classifiers = True



import typing as ty
from .decorators import classproperty
from .exceptions import FormatDefinitionError


class Classifier:
    """Base class for all file-format "classifiers", including datatypes and abstract
    types"""

    @classproperty  # type: ignore[arg-type]
    def type_name(cls) -> str:
        """Name of type to be used in __repr__. Defined here so it can be overridden"""
        return cls.__name__  # type: ignore

    @classproperty  # type: ignore[arg-type]
    def namespace(cls) -> ty.Optional[str]:
        """The "namespace" the format belongs to under the "fileformats" umbrella
        namespace"""
        module_parts = cls.__module__.split(".")
        if module_parts[0] != "fileformats":
            raise FormatDefinitionError(
                f"Cannot create reversible MIME type for {cls} as it is not in the "
                "fileformats namespace"
            )
        return module_parts[1].replace("_", "-")

    def dummy(self) -> float:

        i: int = 0

        return i



import io
from pathlib import Path
import typing as ty
from fileformats.core.fileset import FileSet
from fileformats.core.exceptions import (
    FormatMismatchError,
    UnconstrainedExtensionException,
)
from fileformats.core.decorators import (
    validated_property,
    classproperty,
    mtime_cached_property,
)
from .fsobject import FsObject


class File(FsObject):
    """Generic file type"""

    @validated_property
    def fspath(self) -> Path:
        fspath = self.select_by_ext()
        if not fspath:
            raise FormatMismatchError(
                f"No paths in {type(self)} ({list(self.fspaths)}) "
                f"match extension {self.ext}"
            )
        if fspath.is_dir():
            # fspath is guaranteed to exist
            raise FormatMismatchError(
                f'Path that matches extension of {type(self)}, "{fspath}", '
                f"is a directory not a file"
            )
        return fspath

    @classproperty  # type: ignore[arg-type]
    def unconstrained(cls) -> bool:
        """Whether the file-format is unconstrained by extension, magic number or another
        constraint"""
        return super().unconstrained and (cls.ext is None or None in cls.alternate_exts)

    def is_dir(self) -> bool:
        return False

    def is_file(self) -> bool:
        return True

    @classmethod
    def copy_ext(
        cls,
        old_path: Path,
        new_path: Path,
        decomposition_mode: FileSet.ExtensionDecomposition = FileSet.ExtensionDecomposition.none,
    ) -> Path:
        """Copy extension from the old path to the new path, ensuring that all
        of the extension is used (e.g. 'my.gz' instead of 'gz')

        Parameters
        ----------
        old_path: Path or str
            The path from which to copy the extension from
        new_path: Path or str
            The path to append the extension to
        decomposition_mode : FileSet.ExtensionDecomposition, optional
            if the file doesn't have an explicit extension, how to interpret "." within
            the filename

        Returns
        -------
        Path
            The new path with the copied extension
        """
        if not cls.matching_exts([old_path], [cls.ext]):
            raise FormatMismatchError(
                f"Extension of old path ('{str(old_path)}') does not match that "
                f"of file, '{cls.ext}'"
            )
        suffix = (
            cls.ext
            if cls.ext
            else cls.decompose_fspath(old_path, mode=decomposition_mode)[-1]
        )
        return Path(new_path).with_suffix(suffix)

    @mtime_cached_property
    def raw_contents(self) -> ty.Union[str, bytes]:
        return self.read_contents()

    def open(
        self,
        mode: str = "r",
        buffering: int = -1,
        encoding: ty.Optional[str] = None,
        errors: ty.Optional[str] = None,
        newline: ty.Optional[str] = None,
    ) -> ty.Union[ty.IO[str], ty.IO[bytes]]:
        """Open a I/O stream to the file"""
        return self.fspath.open(
            mode=mode,
            buffering=buffering,
            encoding=encoding,
            errors=errors,
            newline=newline,
        )

    def read_contents(
        self, size: ty.Optional[int] = None, offset: int = 0
    ) -> ty.Union[str, bytes]:
        with self.open("rb" if getattr(self, "binary", True) else "r") as f:
            if offset:
                f.seek(offset, (io.SEEK_SET if offset >= 0 else io.SEEK_END))
            contents = f.read(size) if size else f.read()
        return contents

    @property
    def actual_ext(self) -> str:
        "The actual file extension (out of the primary  and alternate extensions possible)"
        constrained_exts = [
            e for e in self.possible_exts if e is not None
        ]  # strip out unconstrained
        matching = [e for e in constrained_exts if self.fspath.name.endswith(e)]
        if not matching:
            raise UnconstrainedExtensionException(
                f"Cannot determine actual extension of {self.fspath}, as it doesn't "
                f"match any of the defined extensions {constrained_exts} "
                "(i.e. matches the None extension)"
            )
        # Return the longest matching extension, useful for optional extensions
        return sorted(matching, key=len)[-1]  # type: ignore[no-any-return]

    @property
    def stem(self) -> str:
        if self.actual_ext:
            stem = self.fspath.name[: -len(self.actual_ext)]
        else:
            stem = self.fspath.name
        return stem

    def read_bytes(self) -> bytes:
        return self.fspath.read_bytes()


class UnicodeFile(File):

    binary = False

    def open(
        self,
        mode: str = "r",
        buffering: int = -1,
        encoding: ty.Optional[str] = None,
        errors: ty.Optional[str] = None,
        newline: ty.Optional[str] = None,
    ) -> ty.IO[str]:
        """Open a I/O stream to the file"""
        if mode.endswith("b"):
            raise ValueError(
                f"Cannot open text file with binary mode '{mode}'. If you need to do this "
                "access the underlying path directly, e.g. my_file.fspath.open(mode='rb')"
            )
        io: ty.IO[str] = super().open(  # type: ignore[assignment]
            mode=mode,
            buffering=buffering,
            encoding=encoding,
            errors=errors,
            newline=newline,
        )
        return io

    def read_contents(self, size: ty.Optional[int] = None, offset: int = 0) -> str:
        contents: str = super().read_contents(size=size, offset=offset)  # type: ignore[assignment]
        return contents

    def read_text(
        self, encoding: ty.Optional[str] = None, errors: ty.Optional[str] = None
    ) -> str:
        return self.fspath.read_text(encoding=encoding, errors=errors)


class BinaryFile(File):

    binary = True

    def open(
        self,
        mode: str = "r",
        buffering: int = -1,
        encoding: ty.Optional[str] = None,
        errors: ty.Optional[str] = None,
        newline: ty.Optional[str] = None,
    ) -> ty.IO[bytes]:
        """Open a I/O stream to the file"""
        if not mode.endswith("b"):
            mode += "b"
        io: ty.IO[bytes] = super().open(  # type: ignore[assignment]
            mode=mode,
            buffering=buffering,
            encoding=encoding,
            errors=errors,
            newline=newline,
        )
        return io

    def read_contents(self, size: ty.Optional[int] = None, offset: int = 0) -> bytes:
        contents: bytes = super().read_contents(size=size, offset=offset)  # type: ignore[assignment]
        return contents



import typing as ty
from pathlib import Path
from fileformats.core.exceptions import FormatMismatchError
from fileformats.core.decorators import (
    validated_property,
    mtime_cached_property,
)
from .fsobject import FsObject
from fileformats.core.fileset import FileSet, FILE_CHUNK_LEN_DEFAULT
from fileformats.core.mixin import WithClassifiers
from fileformats.core.typing import CryptoMethod
from fileformats.core.collection import TypedCollection
from .file import File


class Directory(FsObject):
    """Base directory to be overridden by subtypes that represent directories but don't
    want to inherit content type "qualifers" (i.e. most of them)"""

    content_types: ty.Tuple[ty.Type[FileSet], ...] = ()

    @validated_property
    def fspath(self) -> Path:
        # fspaths are checked for existence with the exception of mock classes
        dirs = [p for p in self.fspaths if not p.is_file()]
        if not dirs:
            raise FormatMismatchError(f"No directory paths provided {repr(self)}")
        if len(dirs) > 1:
            raise FormatMismatchError(
                f"More than one directory path provided {dirs} to {repr(self)}"
            )
        return dirs[0]

    @mtime_cached_property
    def contents(self) -> ty.List[ty.Union[File, "Directory"]]:
        contnts: ty.List[ty.Union[File, Directory]] = []
        for p in self.fspath.iterdir():
            if p.is_dir():
                contnts.append(Directory(p))
            else:
                contnts.append(File(p))
        return contnts

    def is_dir(self) -> bool:
        return True

    def is_file(self) -> bool:
        return False

    def hash_files(
        self,
        crypto: CryptoMethod = None,
        mtime: bool = False,
        chunk_len: int = FILE_CHUNK_LEN_DEFAULT,
        relative_to: ty.Optional[Path] = None,
        ignore_hidden_files: bool = False,
        ignore_hidden_dirs: bool = False,
    ) -> ty.Dict[str, str]:
        if relative_to is None:
            relative_to = self.fspath
        return super().hash_files(
            crypto=crypto,
            mtime=mtime,
            chunk_len=chunk_len,
            relative_to=relative_to,
            ignore_hidden_files=ignore_hidden_files,
            ignore_hidden_dirs=ignore_hidden_dirs,
        )

    @property
    def content_fspaths(self) -> ty.Iterable[Path]:
        return self.fspath.iterdir()

    # Duck-type Path methods

    def __div__(self, other: ty.Union[str, Path]) -> Path:
        return self.fspath / other

    def glob(self, pattern: str) -> ty.Iterator[Path]:
        return self.fspath.glob(pattern)

    def rglob(self, pattern: str) -> ty.Iterator[Path]:
        return self.fspath.rglob(pattern)

    def joinpath(self, other: ty.Union[str, Path]) -> Path:
        return self.fspath.joinpath(other)

    def iterdir(self) -> ty.Iterator[Path]:
        return self.fspath.iterdir()


class TypedDirectory(TypedCollection, Directory):  # type: ignore[misc]
    """Directory that must contain a specific set of content types. Only files that match
    the content types will be considered as contents in the `contents` property.

    Class Attributes
    ----------------
    content_types: ty.Tuple[FileSet, ...]
        the content types that are expected to be found within the directory
    """

    @property
    def content_fspaths(self) -> ty.Iterable[Path]:
        return self.fspath.iterdir()


class DirectoryOf(WithClassifiers, TypedDirectory):  # type: ignore[misc]
    """Generic directory classified by the formats of its contents"""

    # WithClassifiers-required class attrs
    classifiers_attr_name = "content_types"
    allowed_classifiers = (FileSet,)
    allow_optional_classifiers = True
    generically_classifiable = True



from .core import __version__  # noqa
from .core.mixin import WithMagicNumber
from fileformats.generic import BinaryFile


class Video(BinaryFile):
    "Base class for audio file formats"
    binary = True
    # iana_mime = None
    pass


class Mp4(Video):
    """"""

    iana_mime = "video/mp4"
    ext = ".mp4"
    alternate_exts = (".mpg4",)


class Webm(Video):
    ext = ".webm"
    iana_mime = "video/webm"


class Quicktime(Video):
    ext = ".mov"
    alternate_exts = (".qt",)
    iana_mime = "video/quicktime"


class Ogg(WithMagicNumber, Video):
    """"""

    iana_mime = "video/ogg"
    ext = ".ogv"
    magic_number = b"OggS"


class _1d_interleaved_parityfec(Video):
    """"""

    iana_mime = "video/1d-interleaved-parityfec"
    ext = None


class _3gpp(Video):
    """

    TODO: None.  However, the file-type box
    must occur first in the file, and
    MUST contain a 3GPP brand in its
    compatible brands list."""

    iana_mime = "video/3gpp"
    ext = ".3gp"
    alternate_exts = (".3gpp",)


class _3gpp2(Video):
    """Multi-media

    TODO: the file-type box must occur first in the file
    and MUST contain a 3GPP2 brand in its compatible brands list."""

    iana_mime = "video/3gpp2"
    ext = None


class _3gpp_tt(Video):
    """"""

    iana_mime = "video/3gpp-tt"
    ext = None


class Av1(Video):
    """"""

    iana_mime = "video/AV1"
    ext = None


class Bmpeg(Video):
    """"""

    iana_mime = "video/BMPEG"
    ext = None


class Bt656(Video):
    """"""

    iana_mime = "video/BT656"
    ext = None


class Celb(Video):
    """"""

    iana_mime = "video/CelB"
    ext = None


class Dv(Video):
    """Audio and video streaming and
    conferencing tools."""

    iana_mime = "video/DV"
    ext = None


class Encaprtp(Video):
    """"""

    iana_mime = "video/encaprtp"
    ext = None


class Example(Video):
    """"""

    iana_mime = "video/example"
    ext = None


class Ffv1(Video):
    """Any application that requires the transport of  lossless video can use this media type. Some examples are, but not limited to, screen  recording, scientific imaging, and digital video preservation."""

    iana_mime = "video/FFV1"
    ext = None


class Flexfec(Video):
    """Multimedia applications that want to improve resiliency against packet loss by sending redundant data in addition to the source media."""

    iana_mime = "video/flexfec"
    ext = None


class H261(Video):
    """Audio and video streaming and conferencing applications."""

    iana_mime = "video/H261"
    ext = None


class H263(Video):
    """"""

    iana_mime = "video/H263"
    ext = None


class H263_1998(Video):
    """"""

    iana_mime = "video/H263-1998"
    ext = None


class H263_2000(Video):
    """"""

    iana_mime = "video/H263-2000"
    ext = None


class H264(Video):
    """"""

    iana_mime = "video/H264"
    ext = None


class H264_rcdo(Video):
    """"""

    iana_mime = "video/H264-RCDO"
    ext = None


class H264_svc(Video):
    """"""

    iana_mime = "video/H264-SVC"
    ext = None


class H265(Video):
    """"""

    iana_mime = "video/H265"
    ext = None


class H266(Video):
    """Any application that relies
    on VVC-based video services over RTP"""

    iana_mime = "video/H266"
    ext = None


class Iso_Segment(Video):
    """"""

    iana_mime = "video/iso.segment"
    ext = ".m4s"


class Jpeg(Video):
    """"""

    iana_mime = "video/JPEG"
    ext = None


class Jpeg2000(Video):
    """video streaming and communication"""

    iana_mime = "video/jpeg2000"
    ext = None


class Jxsv(Video):
    """Any application that transmits video over RTP (like SMPTE ST
    2110)."""

    iana_mime = "video/jxsv"
    ext = None


class Mj2(WithMagicNumber, Video):
    """"""

    iana_mime = "video/mj2"
    ext = ".mj2"
    alternate_exts = (".mjp2",)
    magic_number = "0000000C6A5020200D0A870A"


class Mp1s(Video):
    """"""

    iana_mime = "video/MP1S"
    ext = None


class Mp2p(Video):
    """"""

    iana_mime = "video/MP2P"
    ext = None


class Mp2t(Video):
    """"""

    iana_mime = "video/MP2T"
    ext = None


class Mp4v_es(Video):
    """"""

    iana_mime = "video/MP4V-ES"
    ext = None


class Mpv(Video):
    """"""

    iana_mime = "video/MPV"
    ext = None


class Mpeg4_generic(Video):
    """"""

    iana_mime = "video/mpeg4-generic"
    ext = None


class Nv(Video):
    """Audio and video streaming and conferencing tools."""

    iana_mime = "video/nv"
    ext = None


class Parityfec(Video):
    """"""

    iana_mime = "video/parityfec"
    ext = None


class Pointer(Video):
    """"""

    iana_mime = "video/pointer"
    ext = None


class Raptorfec(Video):
    """Real-time multimedia applications like video streaming, audio streaming, and video conferencing."""

    iana_mime = "video/raptorfec"
    ext = None


class Raw(Video):
    """"""

    iana_mime = "video/raw"
    ext = None


class RtpEncAescm128(Video):
    """"""

    iana_mime = "video/rtp-enc-aescm128"
    ext = None


class Rtploopback(Video):
    """"""

    iana_mime = "video/rtploopback"
    ext = None


class Rtx(Video):
    """"""

    iana_mime = "video/rtx"
    ext = None


class Scip(Video):
    """"""

    iana_mime = "video/scip"
    ext = None


class Smpte291(Video):
    """Devices that stream real-time professional video, especially those that interoperate with legacy serial digital interfaces (SDI)."""

    iana_mime = "video/smpte291"
    ext = None


class Smpte292m(Video):
    """"""

    iana_mime = "video/SMPTE292M"
    ext = None


class Ulpfec(Video):
    """Multimedia applications that seek to improve resiliency to loss by sending additional data with the media stream."""

    iana_mime = "video/ulpfec"
    ext = None


class Vc1(Video):
    """Multimedia streaming and conferencing tools."""

    iana_mime = "video/vc1"
    ext = None


class Vc2(Video):
    """Video Communication."""

    iana_mime = "video/vc2"
    ext = None


class Vp8(Video):
    """"""

    iana_mime = "video/VP8"
    ext = None


class Vp9(Video):
    """"""

    iana_mime = "video/VP9"
    ext = None


__all__ = [
    "Video",
    "Mp4",
    "Webm",
    "Quicktime",
    "Ogg",
    "_1d_interleaved_parityfec",
    "_3gpp",
    "_3gpp2",
    "_3gpp_tt",
    "Av1",
    "Bmpeg",
    "Bt656",
    "Celb",
    "Dv",
    "Encaprtp",
    "Example",
    "Ffv1",
    "Flexfec",
    "H261",
    "H263",
    "H263_1998",
    "H263_2000",
    "H264",
    "H264_rcdo",
    "H264_svc",
    "H265",
    "H266",
    "Iso_Segment",
    "Jpeg",
    "Jpeg2000",
    "Jxsv",
    "Mj2",
    "Mp1s",
    "Mp2p",
    "Mp2t",
    "Mp4v_es",
    "Mpv",
    "Mpeg4_generic",
    "Nv",
    "Parityfec",
    "Pointer",
    "Raptorfec",
    "Raw",
    "RtpEncAescm128",
    "Rtploopback",
    "Rtx",
    "Scip",
    "Smpte291",
    "Smpte292m",
    "Ulpfec",
    "Vc1",
    "Vc2",
    "Vp8",
    "Vp9",
]



import os
import typing as ty
import itertools
from pathlib import Path
from fileformats.core import FileSet, extra_implementation
from fileformats.core import SampleFileGenerator
from .fsobject import FsObject
from .file import File
from .set import TypedSet
from .directory import Directory, TypedDirectory


# Methods to generate sample files, typically used in testing
FILE_FILL_LENGTH = 256


@extra_implementation(FileSet.generate_sample_data)
def fsobject_generate_sample_data(
    fsobject: FsObject,
    generator: SampleFileGenerator,
) -> ty.List[Path]:
    return [generator.generate(File, fill=FILE_FILL_LENGTH)]


@extra_implementation(FileSet.generate_sample_data)
def file_generate_sample_data(
    file: File,
    generator: SampleFileGenerator,
) -> ty.List[Path]:
    contents = None
    if getattr(file, "binary", False):
        if hasattr(file, "magic_number"):
            contents = generator.generate_contents(binary=True)
            magic_number: bytes
            if isinstance(file.magic_number, str):
                magic_number = bytes.fromhex(file.magic_number)
            else:
                magic_number = file.magic_number
            offset = getattr(file, "magic_number_offset", 0)
            if offset < 0:
                postamble = os.urandom(-(len(magic_number) + offset))
                contents += magic_number + postamble  # type: ignore[operator]
            else:
                preamble = generator.generate_contents(binary=True, fill=offset)
                contents = preamble + magic_number + contents  # type: ignore[operator]
        elif hasattr(file, "magic_pattern"):
            raise NotImplementedError(
                "Sampling of magic version file types is not implemented yet"
            )
    elif hasattr(file, "magic_number"):
        raise NotImplementedError(
            f"Sampling of non-binary magic-number file types ({type(file)}) is not implemented yet"
        )
    fspaths = [generator.generate(file, contents=contents, fill=FILE_FILL_LENGTH)]
    if hasattr(file, "header_type"):
        fspaths.extend(file.header_type.sample_data(generator))
    if hasattr(file, "side_car_types"):
        for side_car_type in file.side_car_types:
            fspaths.extend(side_car_type.sample_data(generator))
    return fspaths


@extra_implementation(FileSet.generate_sample_data)
def directory_generate_sample_data(
    directory: Directory,
    generator: SampleFileGenerator,
) -> ty.List[Path]:
    a_dir = generator.generate_fspath(Directory)
    a_dir.mkdir()
    File.sample_data(
        generator.child(dest_dir=a_dir)
    )  # Add a sample file for good measure
    return [a_dir]


@extra_implementation(FileSet.generate_sample_data)
def directory_containing_generate_sample_data(
    directory: TypedDirectory,
    generator: SampleFileGenerator,
) -> ty.List[Path]:
    a_dir = generator.generate_fspath(TypedDirectory)
    a_dir.mkdir()
    for tp in directory.potential_content_types:
        tp.sample_data(generator.child(dest_dir=a_dir))
    return [a_dir]


@extra_implementation(FileSet.generate_sample_data)
def set_of_sample_data(
    set_of: TypedSet,
    generator: SampleFileGenerator,
) -> ty.List[Path]:
    return list(
        itertools.chain(
            *(
                tp.sample_data(generator.child())
                for tp in set_of.potential_content_types
            )
        )
    )



from .core import __version__
from .core.mixin import WithMagicNumber
from fileformats.generic import BinaryFile, File
from fileformats.application import Json, Xml


class Model(File):
    ...  # noqa: E701


class _3mf(Model):
    """3D modeling and slicing  software"""

    iana_mime = "model/3mf"
    ext = ".3mf"


class E57(Model):
    """"""

    iana_mime = "model/e57"
    ext = None


class GltfBinary(Model, BinaryFile):
    """"""

    iana_mime = "model/gltf-binary"
    ext = None


class Gltf__Json(Model, Json):
    """"""

    iana_mime = "model/gltf+json"
    ext = ".gltf"


class Jt(Model):
    """"""

    iana_mime = "model/JT"
    ext = ".jt"


class Iges(Model):
    """"""

    iana_mime = "model/iges"
    ext = None


class Mtl(Model):
    """"""

    iana_mime = "model/mtl"
    ext = ".mtl"


class Obj(Model):
    """"""

    iana_mime = "model/obj"
    ext = ".obj"


class Prc(WithMagicNumber, BinaryFile):
    """"""

    iana_mime = "model/prc"
    ext = ".prc"
    magic_number = "505243"


class Step(Model):
    """"""

    iana_mime = "model/step"
    ext = ".p21"
    alternate_exts = (".stp", ".step", ".stpnc", ".210")


class Step__Xml(Model):
    """"""

    iana_mime = "model/step+xml"
    ext = ".stpx"


class Step__Zip(Model):
    """"""

    iana_mime = "model/step+zip"
    ext = ".stpz"


class StepXml__Zip(Model):
    """"""

    iana_mime = "model/step-xml+zip"
    ext = ".stpxz"


class Stl(Model):
    """"""

    iana_mime = "model/stl"
    ext = ".stl"


class U3d(WithMagicNumber, BinaryFile):
    """"""

    iana_mime = "model/u3d"
    ext = ".u3d"
    magic_number = "55334400"


class X3d_vrml(WithMagicNumber, BinaryFile):
    """"""

    iana_mime = "model/x3d-vrml"
    ext = ".x3dv"
    magic_number = b"#X3D"


class X3d__Fastinfoset(Model):
    """"""

    iana_mime = "model/x3d+fastinfoset"
    ext = ".x3db"


class X3d__Xml(Model, Xml):
    """"""

    iana_mime = "model/x3d+xml"
    ext = ".x3d"


__all__ = [
    "__version__",
    "Model",
    "_3mf",
    "E57",
    "GltfBinary",
    "Gltf__Json",
    "Jt",
    "Iges",
    "Mtl",
    "Obj",
    "Prc",
    "Step",
    "Step__Xml",
    "Step__Zip",
    "StepXml__Zip",
    "Stl",
    "U3d",
    "X3d_vrml",
    "X3d__Fastinfoset",
    "X3d__Xml",
]



class FileFormatsError(Exception):
    "Base exception class"


class FormatDefinitionError(FileFormatsError):
    "When the file-formats class hasn't been properly defined"


class FormatMismatchError(TypeError, FileFormatsError):
    "Provided paths/values do not match the specified file format"


class UnsatisfiableCopyModeError(FileFormatsError):
    "Error copying files"


class FormatConversionError(FileFormatsError):
    "No converters exist between formats"


class FormatRecognitionError(KeyError, FileFormatsError):
    "Did not find a format class corresponding to a MIME, or MIME-like, type string"


class UnconstrainedExtensionException(FileFormatsError):
    "Did not find a format class corresponding to a MIME, or MIME-like, type string"


class FileFormatsExtrasError(FileFormatsError):
    """If there is an "extra" hook in the datatype class but no methods have been
    registered on it"""


class FileFormatsExtrasPkgUninstalledError(FileFormatsExtrasError):
    """If there is an "extra" hook in the datatype class and a extras package on PyPI
    but it hasn't been installed"""


class FileFormatsExtrasPkgNotCheckedError(FileFormatsExtrasError):
    """If there is an "extra" hook in the datatype class and a extras package on PyPI
    but it hasn't been installed"""



import typing as ty
from fileformats.generic import BinaryFile
from .base import Image
from fileformats.core.mixin import WithMagicNumber


class Aces(WithMagicNumber, BinaryFile, Image):
    """

    TODO: or 76 2F 31 01 02 04 00 00 The attribute acesImageContainerFlag has value 1"""

    iana_mime = "image/aces"
    ext = ".exr"
    magic_number = "762f310102000000"


class Apng(WithMagicNumber, BinaryFile, Image):
    """"""

    iana_mime = "image/apng"
    ext = ".apng"
    magic_number = "89504e470d0a1a0a"


class Avci(Image, BinaryFile):
    """"""

    iana_mime = "image/avci"
    ext = ".avci"


class Avcs(Image, BinaryFile):
    """"""

    iana_mime = "image/avcs"
    ext = ".avcs"


class Avif(Image, BinaryFile):
    """"""

    iana_mime = "image/avif"
    ext = ".avif"
    alternative_exts = (".heif", ".heifs", ".hif")


class Cgm(Image, BinaryFile):
    """"""

    iana_mime = "image/cgm"
    ext = None


class DicomRle(Image, BinaryFile):
    """"""

    iana_mime = "image/dicom-rle"
    ext = ".drle"


class Dpx(WithMagicNumber, BinaryFile, Image):
    """

    TODO: 0x53445058 (most-significant byte first
    file) or 58 50"""

    iana_mime = "image/dpx"
    ext = ".dpx"
    magic_number = "53445058"


class Emf(Image, BinaryFile):
    """Office productivity applications; clip art applications; desktop
    publishing applications; some web browsers (e.g., Internet
    Explorer).

    TODO: 0x01000000 (little-endian DWORD 0x00000001), corresponding to
    the EMR_HEADER Type field.
    The next field (EMR_HEADER Size) should be at least 88 (little-
    endian DWORD 0x00000050)."""

    iana_mime = "image/emf"
    ext = ".emf"


class Fits(WithMagicNumber, BinaryFile, Image):
    """There are many astronomical image viewing and data reduction

    A FITS file described with the media type "image/fits" SHOULD have a PHDU with
    positive integer values for the NAXIS and NAXISn keywords, and hence SHOULD contain
    at least one pixel. Files with 4 or more non-degenerate axes (NAXISn>1) SHOULD be
    described as "application/fits", not as "image/fits". (In rare cases it may be
    appropriate to describe a NULL image -- a dataless container for FITS keywords,
    with NAXIS=0 or NAXISn=0 -- or an image with 4+ non- degenerate axes as "image/fits"
    but this usage is discouraged because such files may confuse simple image viewer
    applications.)  FITS files declared as "image/fits" MAY also have one or more
    conforming XHDUs following their PHDUs. These extension HDUs MAY contain standard,
    non-linear, world coordinate system (WCS) information in the form of tables or images.
    The extension HDUs MAY also contain other, non-standard metadata pertaining to the
    image in the PHDU in the forms of keywords and tables.  A FITS file described with
    the media type "image/fits" SHOULD be principally intended to communicate the single
    data array in the PHDU. This means that "image/fits" SHOULD NOT be applied to FITS
    files containing MEF (multi-exposure-frame) mosaic images. Also, random groups files
    MUST be described as "application/fits" and not as "image/fits".  A FITS file described
    with the media type "image/fits" is also valid as a file of media type "application/fits".
    The choice of classification depends on the context and intended usage.

        TODO: "SIMPLE = T"  Jeff Uphoff of the National Radio Astronomy Observatory (NRAO)
        has contributed database entries for the magic number file which is used by the
        Unix "file" command. Magic number files with these entries are distributed with a
        variety of Unix-like operating systems. In addition to recognizing a FITS file using
        the string given above, the Uphoff entries also recognize the data type of the
        pixels in the PHDU.
    """

    iana_mime = "image/fits"
    ext = ".fits"
    alternative_exts = (".fit", ".fts", None)
    magic_number = b"SIMPLE = T"


class G3fax(Image, BinaryFile):
    """"""

    iana_mime = "image/g3fax"
    ext: ty.Optional[str] = None


class Heic(Image, BinaryFile):
    """"""

    iana_mime = "image/heic"
    ext = ".heic"


class HeicSequence(Image, BinaryFile):
    """"""

    iana_mime = "image/heic-sequence"
    ext = ".heics"


class Heif(Image, BinaryFile):
    """"""

    iana_mime = "image/heif"
    ext = ".heif"


class HeifSequence(Image, BinaryFile):
    """"""

    iana_mime = "image/heif-sequence"
    ext = ".heifs"


class Hej2k(Image, BinaryFile):
    """"""

    iana_mime = "image/hej2k"
    ext = ".hej2"


class Hsj2(Image, BinaryFile):
    """"""

    iana_mime = "image/hsj2"
    ext = ".hsj2"


class J2c(WithMagicNumber, BinaryFile, Image):
    """"""

    iana_mime = "image/j2c"
    ext = ".j2c"
    alternative_exts = (".J2C", ".j2k", ".J2K")
    magic_number = "ff4fff51"


class Jls(Image, BinaryFile):
    """"""

    iana_mime = "image/jls"
    ext = ".jls"


class Jp2(WithMagicNumber, BinaryFile, Image):
    """"""

    iana_mime = "image/jp2"
    ext = ".jp2"
    alternative_exts = (".jpg2",)
    magic_number = "0000000c6a5020200d0a870a"


class Jph(Image, BinaryFile):
    """

    TODO: See Section 4.4 of RFC 3745"""

    iana_mime = "image/jph"
    ext = ".jph"


class Jphc(WithMagicNumber, BinaryFile, Image):
    """"""

    iana_mime = "image/jphc"
    ext = ".jhc"
    magic_number = "ff4fff51"


class Jpm(WithMagicNumber, BinaryFile, Image):
    """"""

    iana_mime = "image/jpm"
    ext = ".jpm"
    alternative_exts = (".jpgm",)
    magic_number = "0000000c6a5020200d0a870a"


class Jpx(WithMagicNumber, BinaryFile, Image):
    """"""

    iana_mime = "image/jpx"
    ext = ".jpf"
    magic_number = "0000000c6a5020200d0a870a"


class Jxr(WithMagicNumber, BinaryFile, Image):
    """Imaging, fax, messaging and multimedia.

    TODO: Data begins with a FILE_HEADER( ) data structure, which begins  with a
    FIXED_FILE_HEADER_II_2BYTES field equal to 0x4949, followed by a
    FIXED_FILE_HEADER_0XBC_BYTE field equal to 0xBC, followed by a FILE_VERSION_ID which
    is equal to 1 for the current version of the Recommendation and International
    Standard (with other values reserved for future use, as modified in additional
    parts or amendments, by ITU-T or ISO/IEC).  Within the payload data, JPEG XR
    IMAGE_HEADER( ) data structures begin with a GDI_SIGNATURE, which is a 64-bit
    syntax element that has the value 0x574D50484F544F00 that corresponds to "WMPHOTO"
    using the UTF-8 character set encoding specified in Annex D of ISO/IEC 10646,
    followed by a byte equal to 0.
    """

    iana_mime = "image/jxr"
    ext = ".jxr"
    magic_number = "4949bc"


class Jxra(Image, BinaryFile):
    """"""

    iana_mime = "image/jxrA"
    ext = ".jxra"


class Jxrs(Image, BinaryFile):
    """"""

    iana_mime = "image/jxrS"
    ext = ".jxrs"


class Jxs(WithMagicNumber, BinaryFile, Image):
    """"""

    iana_mime = "image/jxs"
    ext = ".jxs"
    magic_number = "0000000c4a5853200d0a870a"


class Jxsc(WithMagicNumber, BinaryFile, Image):
    """"""

    iana_mime = "image/jxsc"
    ext = ".jxsc"
    magic_number = "ff10ff50"


class Jxsi(Image, BinaryFile):
    """"""

    iana_mime = "image/jxsi"
    ext = ".jxsi"


class Jxss(Image, BinaryFile):
    """"""

    iana_mime = "image/jxss"
    ext = ".jxss"


class Ktx(WithMagicNumber, BinaryFile, Image):
    """Currently only etcpack. It is anticipated it will be widely used by applications
    built on top of the OpenGL family of standards, OpenGL, OpenGL ES and WebGL, as the
    means of delivering texture data."""

    iana_mime = "image/ktx"
    ext = ".ktx"
    magic_number = "ab4b5458203131bb0d0a1a0a"


class Ktx2(WithMagicNumber, BinaryFile, Image):
    """"""

    iana_mime = "image/ktx2"
    ext = ".ktx2"
    magic_number = "ab4b5458203230bb0d0a1a0a"


class Naplps(Image, BinaryFile):
    """"""

    iana_mime = "image/naplps"
    ext = None


class Prs_Btif(Image, BinaryFile):
    """

    TODO: Since BTIF files are an extension of TIFF 6.0, the magic numbers
    are exactly the same as for TIFF 6.0.  The BTIF file
    specification does not introduce any unique magic numbers
    besides what's already present in TIFF 6.0.  (See attached
    specification)"""

    iana_mime = "image/prs.btif"
    ext = ".btif"
    alternative_exts = (".btf",)


class Prs_Pti(Image, BinaryFile):
    """"""

    iana_mime = "image/prs.pti"
    ext = ".pti"


class PwgRaster(WithMagicNumber, BinaryFile, Image):
    """"""

    iana_mime = "image/pwg-raster"
    ext = None
    magic_number = "52615332"


class Svg__Xml(Image, BinaryFile):
    """SVG is used by Web browsers, often in conjunction with HTML; by mobile phones and
    digital cameras, as a format for interchange of graphical assets in desk top publishing,
    for industrial process visualization, display signage, and many other applications
    which require scalable static or interactive graphical capability.

    TODO: Note that the extension ".svgz" is used as an alias for ".svg.gz" [rfc1952]
    i.e. octet streams of type image/svg+xml subsequently compressed with gzip.
    """

    iana_mime = "image/svg+xml"
    ext = ".svg"


class T38(Image, BinaryFile):
    """"""

    iana_mime = "image/t38"
    ext = ".T38"


class TiffFx(Image, BinaryFile):
    """"""

    iana_mime = "image/tiff-fx"
    ext = ".TFX"


class Webp(WithMagicNumber, BinaryFile, Image):
    """Applications that are used to display and process images, especially when smaller
    image file sizes are important.

    TODO: The first 4 bytes are 0x52, 0x49, 0x46, 0x46 ('RIFF'), followed by 4 bytes
    for the RIFF chunk size. The next 7 bytes are 0x57, 0x45, 0x42, 0x50, 0x56, 0x50, 0x38 ('WEBPVP8').
    """

    iana_mime = "image/webp"
    ext = ".webp"
    magic_number = "52494646"


class Wmf(WithMagicNumber, BinaryFile, Image):
    """Office productivity applications; clip art applications; desktop
    publishing applications; some web browsers (e.g., Internet
    Explorer).

    TODO: (little-endian DWORD 0x9AC6CDD7)"""

    iana_mime = "image/wmf"
    ext = ".wmf"
    magic_number = "d7cdc69a"



import operator
from pathlib import Path
import typing as ty
import re
from fileformats.core.exceptions import (
    FormatDefinitionError,
    FormatRecognitionError,
)
import fileformats.core
from .utils import fspaths_converter, add_exc_note


LIST_MIME = "+list-of"
IANA_MIME_TYPE_REGISTRIES = [
    "application",
    "audio",
    "font",
    "image",
    "message",
    "model",
    "multipart",
    "text",
    "video",
]
ALL_STANDARD_TYPE_REGISTRIES = IANA_MIME_TYPE_REGISTRIES + [
    "field",
    "testing",
    "testing_subpackage",
]


def find_matching(
    fspaths: ty.Collection[Path],
    candidates: ty.Optional[ty.Collection[ty.Type["fileformats.core.FileSet"]]] = None,
    standard_only: bool = False,
    include_generic: bool = False,
    skip_unconstrained: bool = True,
) -> ty.List[ty.Type["fileformats.core.FileSet"]]:
    """Detect the corresponding file format from a set of file-system paths

    Parameters
    ----------
    fspaths : list[Path]
        file-system paths to detect the format of
    candidates: sequence[FileSet], optional
        the candidates to select from, by default all file formats
    standard_only : bool, optional
        If you only want to return matches from the "standard" IANA types. Only relevant
        if candidates is None, by default False
    skip_unconstrained : bool, optional
        skip formats that aren't constrained by extension, magic number or another check.
        Only relevant if candidates is None

    Returns
    -------
    list[FileSet]
        the file formats that match the given file-system paths
    """
    import fileformats.core.mixin

    fspaths = fspaths_converter(fspaths)
    matches: ty.List[ty.Type["fileformats.core.FileSet"]] = []
    if candidates is None:
        candidates = fileformats.core.FileSet.all_formats
    for frmt in candidates:
        if skip_unconstrained and frmt.unconstrained:
            continue
        namespace = frmt.namespace
        if (
            frmt.matches(fspaths)
            and (not standard_only or namespace in IANA_MIME_TYPE_REGISTRIES)
            and (include_generic or namespace != "generic")
        ):
            matches.append(frmt)
    return matches


def from_mime(
    mime_str: str,
) -> ty.Union[ty.Type["fileformats.core.DataType"], "ty.Type[ty.Union]"]:
    """Resolves a MIME type (or MIME-like) string into the corresponding type

    Parameters
    ----------
    mime_str : str
        the MIME type, or MIME-like (i.e. using the fileformats namespace scheme
        instead of putting all non-standard types into the 'application' registry), string to
        resolve

    Returns
    -------
    datatype : type
        the resolved datatype

    Raises
    ------
    FormatRecognitionError
        if the MIME string does not correspond to a valid file format class
    """
    if mime_str.endswith(LIST_MIME):
        item_mime = mime_str[: -len(LIST_MIME)]
        if item_mime.startswith("[") and item_mime.endswith("]"):
            item_mime = item_mime[1:-1]
        return ty.List[from_mime(item_mime)]  # type: ignore
    if "," in mime_str:
        return ty.Union.__getitem__(tuple(from_mime(t) for t in mime_str.split(",")))  # type: ignore
    return fileformats.core.DataType.from_mime(mime_str)


def to_mime(
    datatype: ty.Type["fileformats.core.DataType"], official: bool = True
) -> str:
    """Returns the mime-type or mime-like (i.e. using fileformats namespaces instead
    of putting all non-standard types in the 'application' registry) string corresponding
    to the given datatype

    Parameters
    ----------
    datatype : type
        the datatype to get the mime string for
    official : bool
        whether to use the official mime-type instead of mime-like

    Returns
    -------
    mime_str : str
        the MIME type string if `iana=True`, or MIME-like (i.e. using the fileformats
        namespace scheme instead of putting all non-standard types into the 'application'
        registry if not
    """
    origin = ty.get_origin(datatype)
    if official and (origin or datatype.namespace == "field"):
        raise TypeError(
            f"Cannot convert {datatype} to official mime-type as it is not a proper "
            'file-type, please use official=False to convert to "mime-like" string instead'
        )
    if origin is list:
        item_mime = to_mime(ty.get_args(datatype)[0], official=official)
        if "," in item_mime:
            item_mime = "[" + item_mime + "]"
        item_mime += LIST_MIME
        return item_mime
    if origin is ty.Union:
        return ",".join(to_mime(t, official=official) for t in ty.get_args(datatype))
    # Handle case
    if isinstance(datatype, ty.ForwardRef):
        datatype = datatype.__forward_arg__
    if (
        isinstance(datatype, str)
        and datatype.startswith("fileformats.")
        and not official
    ):
        ns, class_name = datatype.split(".")[1:]
        ns = ns.replace("_", "-")
        class_name = to_mime_format_name(class_name)
        return ns + "/" + class_name
    mime: str = datatype.mime_type if official else datatype.mime_like
    if official:
        mime = datatype.mime_type
    else:
        mime = datatype.mime_like
        try:
            from_mime(mime)
        except FormatRecognitionError as e:
            add_exc_note(
                e,
                (
                    f"Cannot create reversible MIME type for {datatype}. Please ensure "
                    "it is imported into a top-level fileformats namespace package "
                    f"'{datatype.namespace}'"
                ),
            )
            raise e
    return mime


def from_paths(
    fspaths: ty.Iterable[Path],
    *candidates: ty.Type["fileformats.core.FileSet"],
    common_ok: bool = False,
    ignore: ty.Optional[str] = None,
    **kwargs: ty.Any,
) -> ty.List["fileformats.core.FileSet"]:
    """Given a list of candidate classes (defaults to all installed in alphabetical order),
    instantiates all possible file-set instances from a collection of file-system paths.

    Note that the order in which the candidates are provided is important as the first
    valid match for each path will be returned.

    Parameters
    ----------
    fspaths : ty.Iterable[Path]
        file-system paths to instantiate file-sets from
    *candidates : tuple[fileformats.core.FileSet]
        the file-set classes to instantiate. If none are provided, then all installed
        filesets will be tried in alphabetical order of their "mime-like" representation.
    common_ok : bool
        whether file-system paths can be used as secondary files in multiple file-sets
    ignore: str, optional
        regular expression pattern for file/directory names to ignore if they aren't
        used in any of the returned file-sets. Any remaining file-paths that are not
        matched by this pattern will cause an error to be raised.
    **kwargs: dict[str, Any]
        keyword arguments passed on to the underlying call to FileSet.from_paths

    Returns
    -------
    list[fileformats.core.FileSet]
        the instantiated file-sets
    """
    if candidates:
        # Unwrap any nested tuples into a flat list of file-setclasses
        unwrapped = []

        def unwrap(candidate: ty.Type["fileformats.core.FileSet"]) -> None:
            if ty.get_origin(candidate) is ty.Union:
                arg: ty.Type["fileformats.core.FileSet"]
                for arg in ty.get_args(candidate):
                    unwrap(arg)
            else:
                unwrapped.append(candidate)

        for candidate in candidates:
            unwrap(candidate)
        candidates = tuple(unwrapped)
        candidates_str = ", ".join(c.mime_like for c in candidates)
    else:
        # Use all installed file-set classes if no candidates are provided, sorted
        # alphabetically to ensure behaviour is consistent between runs
        candidates = tuple(
            sorted(
                fileformats.core.FileSet.subclasses(),
                key=operator.attrgetter("mime_like"),
            )
        )
        candidates_str = "all installed"

    remaining = fspaths
    filesets: ty.List["fileformats.core.FileSet"] = []
    for candidate in candidates:
        fsets, remaining = candidate.from_paths(
            remaining, common_ok=common_ok, **kwargs
        )
        filesets.extend(fsets)
    if ignore:
        ignore_re = re.compile(ignore)
        remaining = [p for p in remaining if not ignore_re.match(p.name)]
    if remaining:
        raise FormatRecognitionError(
            "the following file-system paths were not recognised by any of the "
            f"candidate formats ({candidates_str}):\n"
            + "\n".join(str(p) for p in remaining)
        )
    return filesets


def to_mime_format_name(format_name: str) -> str:
    if "___" in format_name:
        raise FormatDefinitionError(
            f"Cannot convert name of format class {format_name} to mime string as it "
            "contains triple underscore"
        )
    if format_name.startswith("_"):
        format_name = format_name[1:]
    format_name = format_name[0].lower() + format_name[1:]
    format_name = re.sub("__([A-Z])", lambda m: "+" + m.group(1).lower(), format_name)
    format_name = re.sub("_([A-Z])", lambda m: "." + m.group(1).lower(), format_name)
    format_name = re.sub("([A-Z])", lambda m: "-" + m.group(1).lower(), format_name)
    return format_name


def from_mime_format_name(format_name: str) -> str:
    if format_name.startswith("x-"):
        format_name = format_name[2:]
    if re.match(r"^[0-9]", format_name):
        format_name = "_" + format_name
    format_name = format_name.capitalize()
    format_name = re.sub(r"(\.)(\w)", lambda m: "_" + m.group(2).upper(), format_name)
    format_name = re.sub(r"(\+)(\w)", lambda m: "__" + m.group(2).upper(), format_name)
    format_name = re.sub(r"(-)(\w)", lambda m: m.group(2).upper(), format_name)
    return format_name



import os
import typing as ty
import string
from pathlib import Path
import platform
import re
from contextlib import contextmanager
import subprocess as sp
from .utils import logger

PathLike = ty.Union[str, Path]


class FsMountIdentifier:
    """Used to check the mount type that given file paths reside on in order to determine
    features that can be used (e.g. symlinks)"""

    @classmethod
    def symlinks_supported(cls, path: PathLike) -> bool:
        """
        Check whether a file path is on a CIFS filesystem mounted in a POSIX host.

        POSIX hosts are assumed to have the ``mount`` command.

        On Windows, Docker mounts host directories into containers through CIFS
        shares, which has support for Minshall+French symlinks, or text files that
        the CIFS driver exposes to the OS as symlinks.
        We have found that under concurrent access to the filesystem, this feature
        can result in failures to create or read recently-created symlinks,
        leading to inconsistent behavior and ``FileNotFoundError`` errors.

        This check is written to support disabling symlinks on CIFS shares.

        NB: This function and sub-functions are modified from the nipype.utils.filemanip module


        NB: Adapted from https://github.com/nipy/nipype
        """
        return cls.get_mount(path)[1] != "cifs"

    @classmethod
    def on_same_mount(cls, path1: PathLike, path2: PathLike) -> bool:
        """Checks whether two or paths are on the same logical file system"""
        return cls.get_mount(path1)[0] == cls.get_mount(path2)[0]

    @classmethod
    def get_mount(cls, path: PathLike) -> ty.Tuple[Path, str]:
        """Get the mount point for a given file-system path

        Parameters
        ----------
        path: os.PathLike
            the file-system path to identify the mount of

        Returns
        -------
        mount_point: os.PathLike
            the root of the mount the path sits on
        fstype : str
            the type of the file-system (e.g. ext4 or cifs)"""
        strpath = str(Path(path).absolute())
        mount_table = cls.get_mount_table()
        matches = sorted(
            ((Path(p), t) for p, t in mount_table if strpath.startswith(p)),
            key=lambda m: len(str(m[0])),
        )
        if not matches:
            raise ValueError(
                f"Path {strpath} is not on a known mount point:\n{mount_table}"
            )
        # return mount point with longest matching prefix
        return matches[-1]

    @classmethod
    def generate_mount_table(cls) -> ty.List[ty.Tuple[str, str]]:
        """
        Construct a reverse-length-ordered list of mount points that fall under a CIFS mount.

        This precomputation allows efficient checking for whether a given path
        would be on a CIFS filesystem.
        On systems without a ``mount`` command, or with no CIFS mounts, returns an
        empty list.

        """
        if platform.system() == "Windows":
            drive_names = [
                c + ":" for c in string.ascii_uppercase if os.path.exists(c + ":")
            ]
            drives = []
            for drive_name in drive_names:
                result = sp.run(
                    ["fsutil", "fsinfo", "fstype", drive_name],
                    capture_output=True,
                    text=True,
                )
                fstype = result.stdout.strip().split(" ")[-1].lower()
                drives.append((drive_name, fstype))
            return drives
        exit_code, output = sp.getstatusoutput("mount")
        if exit_code != 0:
            raise RuntimeError(
                "Failed to get mount table (exit code {}): {}".format(exit_code, output)
            )
        return cls.parse_mount_table(exit_code, output)

    @classmethod
    def parse_mount_table(
        cls, exit_code: int, output: str
    ) -> ty.List[ty.Tuple[str, str]]:
        """
        Parse the output of ``mount`` to produce (path, fs_type) pairs.

        Separated from _generate_cifs_table to enable testing logic with real
        outputs

        """
        # Linux mount example:  sysfs on /sys type sysfs (rw,nosuid,nodev,noexec)
        #                          <PATH>^^^^      ^^^^^<FSTYPE>
        # OSX mount example:    /dev/disk2 on / (hfs, local, journaled)
        #                               <PATH>^  ^^^<FSTYPE>
        pattern = re.compile(r".*? on (/.*?) (?:type |\()([^\s,\)]+)")

        # Keep line and match for error reporting (match == None on failure)
        # Ignore empty lines
        matches = [(ll, pattern.match(ll)) for ll in output.strip().splitlines() if ll]

        # (path, fstype) tuples, sorted by path length (longest first)
        mounts: ty.List[ty.Tuple[str, str]] = sorted(
            (match.groups() for _, match in matches if match is not None),  # type: ignore
            key=lambda x: len(x[0]),
            reverse=True,
        )

        # Report failures as warnings
        for line, match in matches:
            if match is None:
                logger.debug("Cannot parse mount line: '%s'", line)

        return mounts

    @classmethod
    def get_mount_table(cls) -> ty.List[ty.Tuple[str, str]]:
        if cls._mount_table is None:
            cls._mount_table = cls.generate_mount_table()
        return cls._mount_table

    @classmethod
    @contextmanager
    def patch_table(cls, mount_table: ty.List[ty.Tuple[str, str]]) -> ty.Iterator[None]:
        """Patch the mount table with new values. Used in test routines"""
        orig_table = cls._mount_table
        cls._mount_table = list(mount_table)
        try:
            yield
        finally:
            cls._mount_table = orig_table

    @classmethod
    def get_mtime_resolution(cls, path: PathLike) -> int:
        """Get the mount point for a given file-system path

        Parameters
        ----------
        path: os.PathLike
            the file-system path to identify the mount of

        Returns
        -------
        mount_point: os.PathLike
            the root of the mount the path sits on
        fstype : str
            the type of the file-system (e.g. ext4 or cifs)"""
        mount_point, fstype = cls.get_mount(path)
        try:
            resolution = cls.FS_MAX_MTIME_NS_RESOLUTION[fstype]
        except KeyError:
            resolution = max(cls.FS_MAX_MTIME_NS_RESOLUTION.values())
        return resolution

    _mount_table: ty.Optional[ty.List[ty.Tuple[str, str]]] = None

    # Define a table of file system types and their mtime resolutions (in seconds)
    FS_MAX_MTIME_NS_RESOLUTION: ty.Dict[str, int] = {
        "ext4": int(1e9),  # docs say 1 nanosecond but in found 1 sec often in practice
        "xfs": 1,
        "btrfs": 1,
        "ntfs": 100,
        "hfs": 1,
        "apfs": 1,
        "fat32": int(2e9),  # 2 seconds
        "exfat": int(1e9),  # 1 second
        # Add more file systems and their resolutions as needed
    }



from abc import ABCMeta
import typing as ty
import logging
from .exceptions import FormatDefinitionError
from .classifier import Classifier
from .datatype import DataType

if ty.TYPE_CHECKING:
    from pydra.engine.specs import TaskDef
    import fileformats.core
    from . import mixin

logger = logging.getLogger("fileformats")

T = ty.TypeVar("T")
DT = ty.TypeVar("DT", bound=DataType)


class SubtypeVar:
    """To handle the case where the target format is a placeholder (type-var) defined by
    by its relationship to the source format, e.g.

    AnyFileFormat = FileSet.type_var("AnyFileFormat")

    @converter
    @python.define  # type: ignore[misc]
    def unzip(in_file: Zip[AnyFileFormat], out_file: AnyFileFormat):
        ...
    """

    converters: ty.Dict[ty.Type["fileformats.core.FileSet"], "Converter"] = {}

    @classmethod
    def new(cls, name: str, klass: type) -> "SubtypeVar":
        """Create a new subtype

        Parameters
        ----------
        name : str
            name for the subtype
        klass : ty.Type[DataType]
            the class to sub-type

        Returns
        -------
        SubtypeVar
            a sub-type that is
        """
        return ABCMeta(name, (cls, klass), {"bound": klass})  # type: ignore

    @classmethod
    def __subclasshook__(cls, subclass: type) -> bool:
        if issubclass(subclass, SubtypeVar):
            return issubclass(subclass.bound, cls.bound)  # type: ignore
        return type.__subclasscheck__(cls, subclass)

    @classmethod
    def get_converter_defs(
        cls, source_format: ty.Type["mixin.WithClassifiers"], target_format: type
    ) -> ty.List["Converter"]:
        # check to see whether there are converters from a base class of the source
        # format
        available_converters: ty.List[Converter] = []
        # assert isinstance(source_format, WithClassifiers)
        if source_format.is_classified:
            for template_source_format, converter in cls.converters.items():
                if not issubclass(
                    template_source_format.unclassified, source_format.unclassified  # type: ignore
                ):
                    continue
                assert len(template_source_format.wildcard_classifiers()) == 1  # type: ignore
                non_wildcards = template_source_format.non_wildcard_classifiers()  # type: ignore
                if not non_wildcards.issubset(source_format.classifiers):
                    continue
                from_types = tuple(
                    set(source_format.classifiers).difference(non_wildcards)
                )
                if any(issubclass(q, target_format) for q in from_types):
                    available_converters.append(converter)
        return available_converters

    @classmethod
    def register_converter(
        cls,
        source_format: ty.Type["mixin.WithClassifiers"],
        converter: "Converter",
    ) -> None:
        """Registers a converter task within a class attribute. Called by the
        @fileformats.core.converter decorator.

        Parameters
        ----------
        source_format : type
            the source format to register a converter from
        converter : Converter
            The converter object to register

        Raises
        ------
        FormatConversionError
            if there is already a converter registered between the two types
        """
        # Ensure "converters" dict is defined in the target class and not in a superclass
        if len(source_format.wildcard_classifiers()) > 1:
            raise FormatDefinitionError(
                "Cannot register a conversion to a generic type from a type with more "
                f"than one wildcard {source_format} ({list(source_format.wildcard_classifiers())})"
            )
        prev_registered = [
            f
            for f in cls.converters
            if (
                f.unclassified is source_format.unclassified  # type: ignore
                and f.non_wildcard_classifiers()  # type: ignore
                == source_format.non_wildcard_classifiers()
            )
        ]
        assert len(prev_registered) <= 1
        if prev_registered:
            prev_def = cls.converters[prev_registered[0]]
            # task, task_kwargs, _ = converter_spec
            # prev_task, prev_kwargs = prev_tuple
            if converter.task_def == prev_def.task_def:
                logger.warning(
                    "Ignoring duplicate registrations of the same converter %s",
                    converter.task_def,
                )
                return  # actually the same task but just imported twice for some reason
            generic_type = tuple(prev_def.task_def.wildcard_classifiers())[0]  # type: ignore
            raise FormatDefinitionError(
                f"Cannot register converter from {source_format} to the generic type "
                f"'{generic_type}', {converter.task_def} "
                f"because there is already one registered, {prev_def.task_def}"
            )

        cls.converters[source_format] = converter  # type: ignore


class Converter:
    """Specification of a converter task, including the task callable, its arguments and
    the classifiers"""

    task_def: "TaskDef[ty.Any]"
    classifiers: ty.Tuple[ty.Type[Classifier], ...]
    in_file: str
    out_file: str

    def __init__(
        self,
        task_def: "TaskDef[T]",
        classifiers: ty.Tuple[ty.Type[Classifier], ...] = (),
        in_file: str = "in_file",
        out_file: str = "out_file",
    ):
        self.task_def = task_def
        self.classifiers = classifiers
        self.in_file = in_file
        self.out_file = out_file

    def __eq__(self, other: object) -> bool:
        from pydra.utils.hash import hash_function

        return (
            isinstance(other, Converter)
            and self.task_def == other.task_def
            and hash_function(self.classifiers) == hash_function(other.classifiers)
            and self.in_file == other.in_file
            and self.out_file == other.out_file
        )



import typing as ty
from .utils import (
    fspaths_converter,
)
from .decorators import classproperty
from .typing import FspathsInputType
from .exceptions import (
    FormatDefinitionError,
)

if ty.TYPE_CHECKING:
    from .fileset import FileSet


class MockMixin:
    """Strips out validation methods of a class, allowing it to be mocked in a way that
    still satisfies type-checking"""

    _load_kwargs: ty.Dict[str, ty.Any]

    def __init__(
        self,
        fspaths: FspathsInputType,
        metadata: ty.Union[ty.Dict[str, ty.Any], bool, None] = False,
    ):
        self.fspaths = fspaths_converter(fspaths)
        self._metadata = metadata
        self._load_kwargs = {}

    @classproperty  # type: ignore[arg-type]
    def type_name(cls) -> str:
        return cls.mocked.type_name  # type: ignore[no-any-return]

    def __bytes_repr__(self, cache: ty.Dict[str, ty.Any]) -> ty.Iterable[bytes]:
        yield from (str(fspath).encode() for fspath in self.fspaths)

    @classproperty  # type: ignore[arg-type]
    def mocked(cls) -> "FileSet":
        """The "true" class that the mocked class is based on"""
        return next(c for c in cls.__mro__ if not issubclass(c, MockMixin))  # type: ignore[no-any-return, attr-defined]

    @classproperty  # type: ignore[arg-type]
    def namespace(cls) -> str:
        """The "namespace" the format belongs to under the "fileformats" umbrella
        namespace"""
        mro: ty.Tuple[ty.Type] = cls.__mro__  # type: ignore
        for base in mro:
            if issubclass(base, MockMixin):
                continue
            try:
                return base.namespace  # type: ignore
            except FormatDefinitionError:
                pass
        raise FormatDefinitionError(
            f"None of of the bases classes of {cls} ({mro}) have a valid namespace"
        )



import sys
import typing as ty
from pathlib import Path

if sys.version_info[:2] < (3, 11):
    from typing_extensions import TypeAlias, Self
else:
    from typing import TypeAlias, Self

if ty.TYPE_CHECKING:
    import fileformats.core

CryptoMethod: TypeAlias = ty.Optional[ty.Callable[[], ty.Any]]

FspathsInputType: TypeAlias = ty.Union[  # noqa: F821
    ty.Iterable[ty.Union[str, Path]],
    str,
    Path,
    "fileformats.core.FileSet",
]

PathType: TypeAlias = ty.Union[str, Path]


__all__ = [
    "CryptoMethod",
    "FspathsInputType",
    "PathType",
    "TypeAlias",
    "Self",
]



import os
import sys
import typing as ty
from pathlib import Path
from fileformats.core.fileset import FileSet
from fileformats.core.exceptions import (
    FormatMismatchError,
)
from fileformats.core.decorators import validated_property, classproperty


class FsObject(FileSet, os.PathLike):  # type: ignore
    "Generic file-system object, can be either a file or a directory"

    @validated_property
    def fspath(self) -> Path:
        if len(self.fspaths) > 1:
            fspaths = [str(f) for f in self.fspaths]
            raise FormatMismatchError(
                f"More than one fspath ({fspaths}) provided to FsObject, "
                f"primary path is ambiguous"
            )
        return next(iter(self.fspaths))

    def __str__(self) -> str:
        """Renders the file path as a string so it can be used in templating e.g.
        ``f'cp {fs_object} /tmp'``
        """
        return str(self.fspath)

    def __fspath__(self) -> str:
        """Render to string, so can be treated as any other file-system path, i.e. passed
        to functions like file 'open'"""
        return str(self)

    @property
    def stem(self) -> str:
        return self.fspath.with_suffix("").name

    @classproperty  # type: ignore[arg-type]
    def unconstrained(cls) -> bool:
        """Whether the file-format is unconstrained by extension, magic number or another
        constraint"""
        # We have to subtract `fspath` from required properties as we defined unconstrained
        # file-sets as ones that have more constraints than simply existing
        return not (len(list(cls.validated_properties())) - 1)

    def absolute(self) -> Path:
        return self.fspath

    @property
    def anchor(self) -> str:
        return self.fspath.anchor

    def chmod(self, mode: int, *, follow_symlinks: bool = True) -> None:
        self.fspath.chmod(mode, follow_symlinks=follow_symlinks)  # type: ignore

    @property
    def drive(self) -> str:
        return self.fspath.drive

    def exists(self) -> bool:
        return True

    def group(self) -> ty.Optional[str]:
        if sys.platform == "win32":
            return None
        return self.fspath.group()

    def is_dir(self) -> bool:
        return self.fspath.is_dir()

    def is_file(self) -> bool:
        return self.fspath.is_file()

    @property
    def name(self) -> str:
        return self.fspath.name

    def owner(self) -> ty.Optional[str]:
        if sys.platform == "win32":
            return None
        return self.fspath.owner()

    @property
    def parent(self) -> Path:
        return self.fspath.parent

    @property
    def parents(self) -> ty.Sequence[Path]:
        return self.fspath.parents

    @property
    def parts(self) -> ty.Tuple[str, ...]:
        return self.fspath.parts

    @property
    def root(self) -> str:
        return self.fspath.root

    def stat(self, follow_symlinks: bool = True) -> os.stat_result:
        return self.fspath.stat(follow_symlinks=follow_symlinks)  # type: ignore

    @property
    def suffix(self) -> str:
        return self.fspath.suffix

    @property
    def suffixes(self) -> ty.List[str]:
        return self.fspath.suffixes



from fileformats.generic import UnicodeFile
from fileformats.application import InformalSchema, Xml
from .base import Image


class VectorImage(Image, UnicodeFile):
    pass


class Svg(InformalSchema):
    pass


class Svg__Xml(VectorImage, Xml):
    ext = ".svg"
    iana_mime = "image/svg+xml"
    schema = Svg  # Set Xml schema attribute
    classifiers_attr_name = None  # disable classifiers inherited from Xml



from __future__ import annotations
import typing as ty
from .decorators import classproperty
from .datatype import DataType
from .exceptions import FormatMismatchError


ValueType = ty.TypeVar("ValueType")
PrimitiveType = ty.TypeVar("PrimitiveType")


class Field(ty.Generic[ValueType, PrimitiveType], DataType):
    """Base class for all field formats"""

    value: ValueType
    primitive: ty.Type[PrimitiveType]
    is_field = True
    metadata: None = None  # for duck-typing with FileSet

    def __init__(self, value: ValueType):
        self.value = value

    def __eq__(self, field: object) -> bool:
        return (
            isinstance(field, Field)
            and self.mime_like == field.mime_like
            and self.value == field.value
        )

    def __ne__(self, other: object) -> bool:
        return not self == other

    def __hash__(self) -> int:
        return hash((self.mime_like, self.value))

    def __str__(self) -> str:
        return str(self.value)

    def __repr__(self) -> str:
        return f"{self.type_name}({str(self)})"

    @classproperty  # type: ignore[arg-type]
    def all_fields(
        cls,
    ) -> ty.List[ty.Type["Field[ty.Any, ty.Any]"]]:
        """Iterate over all field formats in fileformats.* namespaces"""
        import fileformats.field  # noqa

        if cls._all_fields is None:
            cls._all_fields = [f for f in Field.subclasses() if getattr(f, "primitive", None) is not None]  # type: ignore
        return cls._all_fields

    def to_primitive(self) -> PrimitiveType:
        return self.primitive(self)  # type: ignore

    @classmethod
    def from_primitive(cls, dtype: type) -> ty.Type[Field[ty.Any, ty.Any]]:
        try:
            datatype: ty.Type[Field[ty.Any, ty.Any]] = next(
                iter(f for f in cls.all_fields if f.primitive is dtype)
            )
        except StopIteration as e:
            field_types_str = ", ".join(t.__name__ for t in cls.all_fields)
            raise FormatMismatchError(
                f"{dtype} doesn't not correspond to a valid fileformats field type "
                f"({field_types_str})"
            ) from e
        return datatype

    _all_fields: ty.Optional[ty.List[ty.Type["Field[ty.Any, ty.Any]"]]] = None
    type: type  # Put this at the bottom to avoid name clash with built-in type
